@@ -24,7 +24,7 @@
 
 ;;; The DEFSTRUCT-DESCRIPTION structure holds compile-time information
 ;;; about a structure type.
-;;; It is defined prior to WRAPPER because WRAPPER-INFO
+;;; It is defined prior to LAYOUT because a LAYOUT-INFO slot
 ;;; is declared to hold a DEFSTRUCT-DESCRIPTION.
 (def!struct (defstruct-description
              (:conc-name dd-)
@@ -137,8 +137,8 @@
              #-sb-xc-host (:pure nil))
   ;; the value to be returned by CLASSOID-NAME.
   (name nil :type symbol)
-  ;; the current WRAPPER for this class, or NIL if none assigned yet
-  (wrapper nil :type (or null wrapper))
+  ;; the current LAYOUT for this class, or NIL if none assigned yet
+  (layout nil :type (or null layout))
   ;; How sure are we that this class won't be redefined?
   ;;   :READ-ONLY = We are committed to not changing the effective
   ;;                slots or superclasses.
@@ -205,11 +205,9 @@
 
 ;;; 32-bit is not done yet. Three slots are still used, instead of two.
 
-#-metaspace
-(progn
-(sb-xc:defstruct (wrapper (:copier nil)
+(sb-xc:defstruct (layout (:copier nil)
                          ;; Parsing DEFSTRUCT uses a temporary layout
-                         (:constructor make-temporary-wrapper
+                         (:constructor make-temporary-layout
                              (clos-hash classoid inherits &aux (invalid nil))))
 
   ;; A packed field containing the DEPTHOID, LENGTH, and FLAGS
@@ -268,6 +266,9 @@
   ;; Could be the generalized function, or a type-specific one
   ;; if the defstruct was compiled in a policy of SPEED 3.
   (equalp-impl #'equalp-err :type (sfunction (t t) boolean) :read-only t)
+  ;; Information for the quicker variant of SLOT-VALUE on STRUCTURE-OBJECT
+  #+64-bit
+  (struct-slot-map nil)
   ;; Information about slots in the class to PCL: this provides fast
   ;; access to slot-definitions and locations by name, etc.
   ;; See MAKE-SLOT-TABLE in pcl/slots-boot.lisp for further details.
@@ -278,44 +279,7 @@
   #-64-bit (id-word3 0 :type word)
   #-64-bit (id-word4 0 :type word)
   #-64-bit (id-word5 0 :type word))
-(declaim (freeze-type wrapper)))
-
-#+metaspace
-(progn
-;;; Separating the pointer and nonpointer slots of LAYOUT satisfies the
-;;; requirement of a certain garbage collector (WIP), but also opens up the
-;;; possibility of absorbing at least one of the vectorish slots into the
-;;; WRAPPER since it could be reallocated with variable size.
-;;; The SLOT-TABLE slot might be a good candidate for trailing tagged slots.
-(sb-xc:defstruct (wrapper (:copier nil) (:constructor %make-wrapper))
-  ;; !!! The FRIEND slot in WRAPPER *MUST* BE FIRST !!! (Wired-in assumption in genesis)
-  (friend nil :type sb-vm:layout)
-  (clos-hash (missing-arg) :type (and fixnum unsigned-byte)) ; redundant
-  (classoid (missing-arg) :type classoid)
-  (inherits #() :type simple-vector)
-  (equalp-impl #'equalp-err :type (sfunction (t t) boolean) :read-only t)
-  (slot-table #(1 nil) :type simple-vector)
-  (%info nil :type (or list defstruct-description))
-  (invalid :uninitialized :type (or cons (member nil t :uninitialized))))
-;;; See #-metaspace structure definition for remarks about each slot.
-;;; LAYOUT points to WRAPPER and vice-versa.
-;;; The most common LAYOUT is 8 words.
-;;; Needing >64 words would be quite unusual - the layout would have
-;;; an enormous depthoid or bitmap or both.
-(sb-xc:defstruct (sb-vm:layout (:copier nil)
-                         ;; Parsing DEFSTRUCT uses a temporary layout
-                         (:constructor %make-temporary-layout (friend clos-hash)))
-  ;; !!! The FRIEND slot in LAYOUT *MUST* BE FIRST !!!
-  (friend nil :type wrapper)
-  (clos-hash (missing-arg) :type (and fixnum unsigned-byte))
-  (flags 0 :type (signed-byte #.sb-vm:n-word-bits))
-  (id-word0 0 :type word)
-  (id-word1 0 :type word)
-  (id-word2 0 :type word)
-  ;; There are zero or more raw words if a type needs to store additional layout-ids,
-  ;; and there are one or more raw words for the GC bitmap.
-  )
-(declaim (freeze-type wrapper sb-vm:layout)))
+(declaim (freeze-type layout))
 
 (declaim (end-block))
 
@@ -330,11 +294,9 @@
 ;;; FLAGS are computed on demand, and not stored.
 #+sb-xc-host
 (progn
-  (defstruct (wrapper (:constructor host-make-wrapper
+  (defstruct (layout (:constructor host-make-layout
                          (id clos-hash classoid
-                          &key ((:info %info)) depthoid inherits length invalid
-                          #+metaspace friend)))
-    #+metaspace (friend)
+                          &key ((:info %info)) depthoid inherits length invalid)))
     (id nil :type (or null fixnum))
     ;; Cross-compiler-only translation from slot index to symbol naming
     ;; the accessor to call. (Since access by position is not a thing)
@@ -350,12 +312,10 @@
     (depthoid -1 :type layout-depthoid)
     (length 0 :type layout-length)
     (%info nil :type (or null defstruct-description)))
-  #+metaspace (defstruct (sb-vm:layout (:constructor %make-layout))
-                friend)
-  (defun make-temporary-wrapper (clos-hash classoid inherits)
-    (host-make-wrapper nil clos-hash classoid :inherits inherits :invalid nil))
-  (defun wrapper-flags (wrapper)
-    (declare (type wrapper wrapper))
+  (defun make-temporary-layout (clos-hash classoid inherits)
+    (host-make-layout nil clos-hash classoid :inherits inherits :invalid nil))
+  (defun layout-flags (layout)
+    (declare (type layout layout))
     (let ((mapping `((structure-object  ,+structure-layout-flag+)
                      (standard-object   ,+pcl-object-layout-flag+)
                      (pathname          ,+pathname-layout-flag+)
@@ -365,10 +325,10 @@
                      (stream            ,+stream-layout-flag+)
                      (sequence          ,+sequence-layout-flag+)))
           (flags 0))
-      (dolist (x (cons wrapper (coerce (wrapper-inherits wrapper) 'list)))
-        (let ((cell (assoc (wrapper-classoid-name x) mapping)))
+      (dolist (x (cons layout (coerce (layout-inherits layout) 'list)))
+        (let ((cell (assoc (layout-classoid-name x) mapping)))
           (when cell (setq flags (logior flags (second cell))))))
-      (let ((dd (wrapper-%info wrapper)))
+      (let ((dd (layout-%info layout)))
         (when (or (logtest flags (logior +pathname-layout-flag+ +condition-layout-flag+))
                   (and (logtest flags +structure-layout-flag+)
                        dd
@@ -378,39 +338,36 @@
       ;; any more complicated than necessary. It is unable to express that
       ;; these funcallable instances can go on pure boxed pages.
       ;; (The trampoline is always an assembler routine, thus ignorable)
-      (when (member (wrapper-classoid-name wrapper)
+      (when (member (layout-classoid-name layout)
                     '(sb-pcl::ctor sb-pcl::%method-function))
         (setf flags (logior flags +strictly-boxed-flag+)))
       flags))
-  (defun wrapper-bitmap (wrapper)
-    (acond ((wrapper-%info wrapper) (dd-bitmap it))
+  (defun layout-bitmap (layout)
+    (acond ((layout-%info layout) (dd-bitmap it))
            ;; Give T a 0 bitmap. It's arbitrary, but when we need some layout
            ;; that has this bitmap we can use the layout of T.
-           ((eq wrapper (find-layout t)) 0)
+           ((eq layout (find-layout t)) 0)
            (t
             +layout-all-tagged+)))
-  (defun %layout-bitmap (layout) (wrapper-bitmap layout))
+  (defun %layout-bitmap (layout) (layout-bitmap layout))
 ) ; end PROGN #+sb-xc-host
 
 (defun equalp-err (a b)
   (bug "EQUALP ~S ~S" a b))
 
-(defmacro name->dd (name)
-  ;; This wants to be a toplevel macrolet but can't be, because the body of
-  ;; the macro (which is run in the host) wouldn't see NAME->DD as a macro
-  ;; when expanding for the target. And it can't be a toplevel FLET because
-  ;; that would demote the two using macros from toplevel.
-  `(find-defstruct-description (cond #-metaspace ((eq ,name 'sb-vm:layout) 'wrapper)
-                                     (t ,name))))
-(defmacro type-dd-length (type-name) (dd-length (name->dd type-name)))
 (defmacro get-dsd-index (type-name slot-name)
   (declare (notinline dsd-index)) ; avoid later inlining failure style-warning
-  (dsd-index (find slot-name (dd-slots (name->dd type-name)) :key #'dsd-name)))
+  (dsd-index (find slot-name
+                   (dd-slots (find-defstruct-description type-name))
+                   :key #'dsd-name)))
 
 ;;; Applicable only if bit-packed (for 64-bit architectures)
 (defmacro pack-layout-flags (depthoid length flags)
   `(logior (ash ,depthoid (+ 32 sb-vm:n-fixnum-tag-bits)) (ash ,length 16) ,flags))
 
+(defmacro type-dd-length (type-name)
+  (dd-length (find-defstruct-description type-name)))
+
 (defconstant layout-id-vector-fixed-capacity 7)
 (defmacro calculate-extra-id-words (depthoid)
   ;; There are 1 or 2 ids per word depending on n-word-bytes.
@@ -420,17 +377,18 @@
   `(ceiling (max 0 (- ,depthoid ,layout-id-vector-fixed-capacity))
             ,(/ sb-vm:n-word-bytes 4)))
 
-(declaim (inline wrapper-info wrapper-dd))
-(defun wrapper-info (wrapper)
-  (let ((info (wrapper-%info wrapper))) (unless (listp info) info)))
-(defun (setf wrapper-info) (newval wrapper)
+(declaim (inline layout-dd layout-info))
+;; Use LAYOUT-DD to read LAYOUT-INFO if you want to assert that it is non-nil.
+(defun layout-dd (layout)
+  (the defstruct-description (layout-%info layout)))
+(defun layout-info (layout)
+  (let ((info (layout-%info layout)))
+    (unless (listp info) info)))
+(defun (setf layout-info) (newval layout)
   ;; The current value must be nil or a defstruct-description,
   ;; otherwise we'd clobber a non-nil slot list.
-  (aver (not (consp (wrapper-%info wrapper))))
-  (setf (wrapper-%info wrapper) newval))
-;; Use WRAPPER-DD to read WRAPPER-INFO and assert that it is non-nil.
-(defun wrapper-dd (wrapper)
-  (the defstruct-description (wrapper-%info wrapper)))
+  (aver (not (consp (layout-%info layout))))
+  (setf (layout-%info layout) newval))
 
 ;;; See the pictures above DD-BITMAP in src/code/defstruct for the details.
 (defconstant standard-gf-primitive-obj-layout-bitmap
@@ -454,13 +412,13 @@
 (progn
 (declaim (inline bitmap-start bitmap-nwords bitmap-all-taggedp))
 (defun bitmap-start (layout)
-  (+ (type-dd-length sb-vm:layout)
+  (+ (type-dd-length layout)
      (calculate-extra-id-words (layout-depthoid layout))))
 (defun bitmap-nwords (layout)
-  (declare (sb-vm:layout layout))
+  (declare (layout layout))
   (- (%instance-length layout)
      (calculate-extra-id-words (layout-depthoid layout))
-     (type-dd-length sb-vm:layout)))
+     (type-dd-length layout)))
 (defun bitmap-all-taggedp (layout)
   ;; All bitmaps have at least 1 word; read that first.
   (and (= (%raw-instance-ref/signed-word layout (bitmap-start layout))
@@ -468,36 +426,28 @@
        ;; Then check that there are no additional words.
        (= (bitmap-nwords layout) 1)))
 
-#+metaspace ; If metaspace, then WRAPPER has no flags; they're in the LAYOUT.
-(defmacro wrapper-flags (x) `(layout-flags (wrapper-friend ,x)))
-
-;; 32-bit has the depthoid as a slot, 64-bit has it is part of FLAGS which are
-;; in the LAYOUT
-#+64-bit
-(defun wrapper-depthoid (wrapper) (layout-depthoid (wrapper-friend wrapper)))
-
 ;;; Extract the bitmap from 1 or more words of bits that have the same format
 ;;; as a BIGNUM - least-significant word first, native bit order within each word,
 ;;; all but the last are unsigned, and the last is signed.
-(defun %layout-bitmap (wrapper)
-  (declare (type wrapper wrapper))
+(defun %layout-bitmap (layout)
+  (declare (type layout layout))
   (do ((accumulator 0)
        (shift 0)
-       (index (bitmap-start wrapper) (1+ index))
-       (last (1- (%instance-length wrapper))))
+       (index (bitmap-start layout) (1+ index))
+       (last (1- (%instance-length layout))))
       ((> index last) accumulator)
     (setf accumulator (logior accumulator
                               (ash (if (= index last)
-                                       (%raw-instance-ref/signed-word wrapper index)
-                                       (%raw-instance-ref/word wrapper index))
+                                       (%raw-instance-ref/signed-word layout index)
+                                       (%raw-instance-ref/word layout index))
                                    shift)))
     (incf shift sb-vm:n-word-bits)))
 
-(defun wrapper-bitmap (wrapper)
-  (declare (type wrapper wrapper))
-  ;; Whenever we call WRAPPER-BITMAP on a structure-object subtype,
+(defun layout-bitmap (layout)
+  (declare (type layout layout))
+  ;; Whenever we call LAYOUT-BITMAP on a structure-object subtype,
   ;; it's supposed to have the INFO slot populated, linking it to a defstruct-description.
-  (acond ((wrapper-info wrapper) (dd-bitmap it))
+  (acond ((layout-info layout) (dd-bitmap it))
          ;; Instances lacking DD-INFO are CLOS objects, which can't generally have
          ;; raw slots, except that funcallable-instances may have 2 raw slots -
          ;; the trampoline and the layout. The trampoline can have a tag, depending
@@ -505,33 +455,33 @@
          ;; In any event, the bitmap is always 1 word, and there are no "extra ID"
          ;; words preceding it.
          (t
-          (aver (not (logtest +structure-layout-flag+ (wrapper-flags wrapper))))
+          (aver (not (logtest +structure-layout-flag+ (layout-flags layout))))
           (the fixnum
-               (%raw-instance-ref/signed-word (wrapper-friend wrapper)
-                                              (type-dd-length sb-vm:layout))))))
+               (%raw-instance-ref/signed-word layout
+                                              (type-dd-length layout))))))
 #+64-bit
-(defmacro wrapper-length (wrapper) ; SETFable
-  `(ldb (byte 16 16) (layout-flags (wrapper-friend ,wrapper))))
+(defmacro layout-length (layout) ; SETFable
+  `(ldb (byte 16 16) (layout-flags ,layout)))
 
 ) ; end PROGN #-sb-xc-host
 
 ;;; True of STANDARD-OBJECT, which include generic functions.
 (declaim (inline layout-for-pcl-obj-p))
-(defun layout-for-pcl-obj-p (wrapper)
-  (declare (type wrapper wrapper))
-  (logtest (wrapper-flags wrapper) +pcl-object-layout-flag+))
+(defun layout-for-pcl-obj-p (layout)
+  (declare (type layout layout))
+  (logtest (layout-flags layout) +pcl-object-layout-flag+))
 
-(defun wrapper-classoid-name (x)
-  (classoid-name (wrapper-classoid x)))
+(defun layout-classoid-name (x)
+  (classoid-name (layout-classoid x)))
 
 ;;;; object types to represent classes
 
 ;;; An UNDEFINED-CLASSOID is a cookie we make up to stick in forward
 ;;; referenced layouts. Users should never see them.
 (defstruct (undefined-classoid
-             (:include classoid)
-             (:constructor !alloc-undefined-classoid (%bits name))
-             (:copier nil)))
+            (:include classoid)
+            (:constructor !alloc-undefined-classoid (%bits name))
+            (:copier nil)))
 
 ;;; BUILT-IN-CLASS is used to represent the standard classes that
 ;;; aren't defined with DEFSTRUCT and other specially implemented
