@@ -14,8 +14,6 @@
 
 (in-package "SB-VM")
 
-(eval-when (:compile-toplevel :load-toplevel :execute)
-
 (defconstant-eqx +float-trap-alist+
     `((:underflow . ,float-underflow-trap-bit)
       (:overflow . ,float-overflow-trap-bit)
@@ -40,13 +38,13 @@
   #'equal)
 
 ;;; Return a mask with all the specified float trap bits set.
-(defun float-trap-mask (names)
-  (reduce #'logior
-          (mapcar (lambda (x)
-                    (or (cdr (assoc x +float-trap-alist+))
-                        (error "unknown float trap kind: ~S" x)))
-                  names)))
-) ; EVAL-WHEN
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (defun float-trap-mask (names)
+    (reduce #'logior
+            (mapcar (lambda (x)
+                      (or (cdr (assoc x +float-trap-alist+))
+                          (error "unknown float trap kind: ~S" x)))
+                    names))))
 
 ;;; interpreter stubs for floating point modes get/setters for
 ;;; some architectures have been removed, as they are implemented
@@ -172,8 +170,8 @@ sets the floating point modes to their current values (and thus is a no-op)."
 ;;; Return true if any of the named traps are currently trapped, false
 ;;; otherwise.
 (defmacro current-float-trap (&rest traps)
-  `(not (zerop (logand ,(dpb (float-trap-mask traps) float-traps-byte 0)
-                       (floating-point-modes)))))
+  `(logtest ,(dpb (float-trap-mask traps) float-traps-byte 0)
+            (floating-point-modes)))
 
 ;;; SIGFPE code to floating-point error
 #-win32
