@@ -25,6 +25,12 @@
 (when (sb-sys:find-dynamic-foreign-symbol-address "show_gc_generation_throughput")
   (setf (extern-alien "show_gc_generation_throughput" int) 1))
 
+#+sbcl ; prevent "illegal to redefine standard type: RATIONAL" etc
+(when (member "SB-XC" (package-nicknames "CL") :test 'string=)
+  (sb-ext:unlock-package "CL")
+  (rename-package "CL" "COMMON-LISP" '("CL"))
+  (sb-ext:lock-package "CL"))
+
 (in-package "SB-COLD")
 
 (defun parse-make-host-parallelism (str)
@@ -280,16 +286,27 @@
              ;; Bind temporarily so that TARGET-FEATUREP and TARGET-PLATFORM-KEYWORD
              ;; can see the tentative list.
              (sb-xc:*features* (funcall customizer default-features))
-             (gc (find-if (lambda (x) (member x '(:cheneygc :gencgc)))
-                          sb-xc:*features*))
+             (gc (intersection '(:cheneygc :gencgc :mark-region-gc)
+                               sb-xc:*features*))
              (arch (target-platform-keyword)))
+        (unless (and gc (not (cdr gc)))
+          (error "Exactly 1 GC implementation needs to be selected"))
+        (setq gc (car gc))
+        ;; all our GCs are generational
+        (when (member gc '(:gencgc :mark-region-gc))
+          (pushnew :generational sb-xc:*features*))
+        (when (eq gc :mark-region-gc)
+          (setq sb-xc:*features* (remove :immobile-space sb-xc:*features*)))
         ;; Win32 conditionally adds :sb-futex in grovel-features.sh
-        (when (target-featurep '(:and :sb-thread (:or :linux :freebsd)))
+        ;; Futexes aren't available in all macos versions, but they are available in
+        ;; all versions that support arm, so always enable them there
+        (when (target-featurep '(:and :sb-thread (:or :linux :freebsd :openbsd (:and :darwin :arm64))))
           (pushnew :sb-futex sb-xc:*features*))
+        (when (target-featurep '(:and :sb-thread :x86-64))
+          (pushnew :system-tlabs sb-xc:*features*))
         (when (target-featurep :immobile-space)
-          (when (member :sb-thread sb-xc:*features*)
-            (pushnew :system-tlabs sb-xc:*features*))
-          (pushnew :compact-instance-header sb-xc:*features*)
+          (when (target-featurep :x86-64)
+            (pushnew :compact-instance-header sb-xc:*features*))
           (pushnew :immobile-code sb-xc:*features*))
         (when (target-featurep :64-bit)
           (push :compact-symbol sb-xc:*features*))
@@ -304,6 +321,17 @@
           ;; Just print something and go on with life.
           (setq sb-xc:*features* (remove :int4-breakpoints sb-xc:*features*))
           (warn "Removed :INT4-BREAKPOINTS from target features"))
+        (when (target-featurep :x86-64)
+          (let ((int3-enable (target-featurep :int3-breakpoints))
+                (int4-enable (target-featurep :int4-breakpoints))
+                (ud2-enable (target-featurep :ud2-breakpoints)))
+            (when (or ud2-enable int4-enable)
+              (setq sb-xc:*features* (remove :int3-breakpoints sb-xc:*features*))
+              (when (and ud2-enable int4-enable)
+                (error "UD2-BREAKPOINTS and INT4-BREAKPOINTS are mutually exclusive choices")))
+            (unless (or int3-enable int4-enable ud2-enable)
+              ;; don't love the name, but couldn't think of a better one
+              (push :sw-int-avoidance sb-xc:*features*))))
         (when (or (target-featurep :arm64)
                   (and (target-featurep :x86-64)
                        (member :sse4 backend-subfeatures)))
@@ -333,17 +361,13 @@
 ;;; combinations now, and provide a description of what the actual
 ;;; failure is (not always obvious from when the build fails).
 (let ((feature-compatibility-tests
-       '(("(and sb-thread (not gencgc))"
-          ":SB-THREAD requires :GENCGC")
-         ("(and sb-safepoint (not sb-thread))" ":SB-SAFEPOINT requires :SB-THREAD")
+       '(("(and sb-safepoint (not sb-thread))" ":SB-SAFEPOINT requires :SB-THREAD")
          ("(and sb-thread (not (or riscv ppc ppc64 x86 x86-64 arm64)))"
           ":SB-THREAD not supported on selected architecture")
          ("(and (not sb-thread) (or arm64 ppc64))"
           "The selected architecture requires :SB-THREAD")
          ("(and gencgc cheneygc)"
           ":GENCGC and :CHENEYGC are incompatible")
-         ("(not (or gencgc cheneygc))"
-          "One of :GENCGC or :CHENEYGC must be enabled")
          ("(and sb-safepoint (not (and (or arm64 x86 x86-64) (or darwin linux win32))))"
           ":SB-SAFEPOINT not supported on selected arch/OS")
          ("(not (or elf mach-o win32))"
@@ -357,18 +381,14 @@
           ;; It sorta kinda works to have both, but there should be no need,
           ;; and it's not really supported.
           "At most one interpreter can be selected")
-         ("(and immobile-space (not x86-64))"
-          ":IMMOBILE-SPACE is supported only on x86-64")
          ("(and compact-instance-header (not immobile-space))"
           ":COMPACT-INSTANCE-HEADER requires :IMMOBILE-SPACE feature")
          ("(and immobile-code (not immobile-space))"
           ":IMMOBILE-CODE requires :IMMOBILE-SPACE feature")
          ("(and immobile-symbols (not immobile-space))"
           ":IMMOBILE-SYMBOLS requires :IMMOBILE-SPACE feature")
-         ("(and system-tlabs (or (not sb-thread) (not immobile-space)))"
-          ;; I don't think it really reqires immobile-space but I haven't tried
-          ;; it and I don't care to support that.
-          ":SYSTEM-TLABS requires SB-THREAD and IMMOBILE-SPACE")
+         ("(and system-tlabs (not sb-thread))"
+          ":SYSTEM-TLABS requires SB-THREAD")
          ("(and sb-futex (not sb-thread))"
           "Can't enable SB-FUTEX on platforms lacking thread support")
          ;; There is still hope to make multithreading on DragonFly x86-64
@@ -752,7 +772,10 @@
              (search "src/code/arena" stem)
              (search "src/code/avltree" stem)
              (search "src/code/brothertree" stem)
-             (search "src/code/debug" stem)
+             (search "src/code/early-classoid" stem)
+             (search "src/code/type-class" stem)
+             (search "src/code/class" stem)
+             (search "src/code/debug" stem) ; also matches debug-{info,int,var-io}
              (search "src/code/early-defmethod" stem)
              (search "src/code/final" stem)
              (search "src/code/format" stem)
@@ -820,7 +843,7 @@
                ((and structure-object (not package))
                 (let ((type-name (string (type-of x))))
                   ;; This "LAYOUT" refers to *our* object, not host-sb-kernel:layout.
-                  (unless (member type-name '("WRAPPER" "LAYOUT" "FLOAT" "COMPLEXNUM")
+                  (unless (member type-name '("LAYOUT" "FLOAT" "COMPLEXNUM")
                                   :test #'string=)
                     ;(Format t "visit a ~/host-sb-ext:print-symbol-with-prefix/~%" (type-of x))
                     ;; This generalizes over any structure. I need it because we
