@@ -1,4 +1,4 @@
-;;;; Type checking of higher order function.
+;;;; Type checking of higher order functions.
 ;;;;
 ;;;; This software is part of the SBCL system. See the README file for
 ;;;; more information.
@@ -9,7 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 (defun assert-function-designator (caller lvars lvar annotation policy)
   (destructuring-bind (args &optional results &rest options) annotation
@@ -31,8 +31,6 @@
                           :deps deps
                           :type type)))
         (when (add-annotation lvar annotation)
-          (loop for lvar in deps
-                do (push annotation (lvar-dependent-annotations lvar)))
           (assert-lvar-type lvar
                             (specifier-type 'function-designator)
                             policy))))))
@@ -42,7 +40,7 @@
      (length (fun-type-optional fun-type))))
 
 (defun callable-dependant-lvars (caller lvars args results)
-  (let ((fun-type (info :function :type caller)))
+  (let ((fun-type (global-ftype caller)))
     (collect ((lvars))
       (let ((arg-position -1)
             (positional-count (fun-type-positional-count fun-type)))
@@ -51,7 +49,8 @@
                    (incf arg-position))
                  (handle-keys (options)
                    (loop for (key value*) on options by #'cddr
-                         for value = (if (eq key :key)
+                         for value = (if (or (eq key :key)
+                                             (eq key :value))
                                          (let ((lvar (getf (nthcdr positional-count lvars) value*)))
                                            (and lvar
                                                 (record-lvar lvar)))
@@ -126,9 +125,9 @@
 ;;; The function should accept
 ;;; (lvar args results &key (unknown-keys boolean) (no-function-conversion boolean) (arg-lvars list-of-lvars))
 (defun map-callable-arguments (function combination)
-  (let* ((comination-name (lvar-fun-name (combination-fun combination) t))
-         (type (info :function :type comination-name))
-         (info (info :function :info comination-name))
+  (let* ((combination-name (lvar-fun-name (combination-fun combination) t))
+         (type (global-ftype combination-name))
+         (info (info :function :info combination-name))
          (annotation (fun-info-annotation info)))
     (when annotation
       (multiple-value-bind (arg-lvars unknown) (resolve-key-args (combination-args combination) type)
@@ -151,7 +150,7 @@
                   (when lvar
                     (call lvar annotation)))))))))
 
-(defun lvar-fun-type (lvar &optional defined-here)
+(defun lvar-fun-type (lvar &optional defined-here asserted-type)
   ;; Handle #'function,  'function and (lambda (x y))
   (let* ((use (principal-lvar-use lvar))
          (lvar-type (lvar-type lvar))
@@ -164,28 +163,65 @@
                                 (node-source-form use))
                                (t
                                 '.anonymous.))))))
+         (asserted t)
          (defined-type (and (global-var-p leaf)
                             (case (leaf-where-from leaf)
                               (:declared
                                (leaf-type leaf))
-                              ((:defined :defined-here)
-                               (if (or (and (defined-fun-p leaf)
-                                            (eq (defined-fun-inlinep leaf) :notinline))
-                                       (and defined-here
-                                            (eq (leaf-where-from leaf) :defined))
-                                       (fun-lexically-notinline-p (leaf-%source-name leaf)
-                                                                  (node-lexenv (lvar-dest lvar))))
-                                   lvar-type
-                                   (proclaimed-ftype (leaf-%source-name leaf))))
+                              (:defined
+                               (cond ((or defined-here
+                                          asserted-type
+                                          (and (defined-fun-p leaf)
+                                               (eq (defined-fun-inlinep leaf) 'notinline))
+                                          (fun-lexically-notinline-p (leaf-%source-name leaf)
+                                                                     (node-lexenv (lvar-dest lvar))))
+                                      (setf asserted nil)
+                                      lvar-type)
+                                     (t
+                                      (global-ftype (leaf-%source-name leaf)))))
+                              (:defined-here
+                               (cond ((or (and (defined-fun-p leaf)
+                                               (eq (defined-fun-inlinep leaf) 'notinline))
+                                          (fun-lexically-notinline-p (leaf-%source-name leaf)
+                                                                     (node-lexenv (lvar-dest lvar))))
+                                      lvar-type)
+                                     (t
+                                      (global-ftype (leaf-%source-name leaf)))))
                               (t
                                (global-var-defined-type leaf)))))
-         (lvar-type (if (and defined-type
-                             (neq defined-type *universal-type*))
-                        defined-type
-                        lvar-type))
+         (entry-fun (if (and (functional-p leaf)
+                             (eq (functional-kind leaf) :external))
+                        (functional-entry-fun leaf)
+                        leaf))
+         (lvar-type (cond ((and defined-type
+                                (neq defined-type *universal-type*))
+                           defined-type)
+                          ((and (functional-p entry-fun)
+                                (fun-type-p (functional-type entry-fun)))
+                           (functional-type entry-fun))
+                          ((and (not (fun-type-p lvar-type))
+                                (lambda-p entry-fun)
+                                (null (lambda-kind entry-fun))
+                                (lambda-tail-set entry-fun))
+                           (make-fun-type :wild-args t
+                                          :returns
+                                          (tail-set-type (lambda-tail-set entry-fun))))
+                          ((and asserted-type
+                                (not (or (constant-lvar-p lvar)
+                                         (constant-p leaf))))
+                           (setf asserted nil)
+                           ;; Don't trust FUNCTION type declarations,
+                           ;; they perform no runtime assertions.
+                           (specifier-type 'function))
+                          (t
+                           (setf asserted nil)
+                           lvar-type)))
          (fun-name (cond ((or (fun-type-p lvar-type)
-                              (functional-p leaf))
-                          (cond ((constant-lvar-p lvar)
+                              (functional-p leaf)
+                              (global-var-p leaf))
+                          (cond ((or (constant-lvar-p lvar)
+                                     ;; A constant may fail some checks in constant-lvar-p
+                                     (constant-p leaf))
                                  (let ((value (lvar-value lvar)))
                                    (etypecase value
                                      #-sb-xc-host
@@ -205,10 +241,12 @@
          (type (cond ((fun-type-p lvar-type)
                       lvar-type)
                      ((symbolp fun-name)
-                      (if (fun-lexically-notinline-p fun-name
-                                                     (node-lexenv (lvar-dest lvar)))
+                      (if (or (fun-lexically-notinline-p fun-name
+                                                         (node-lexenv (lvar-dest lvar)))
+                              (and (neq (info :function :where-from fun-name) :declared)
+                                   asserted-type))
                           lvar-type
-                          (proclaimed-ftype fun-name)))
+                          (global-ftype fun-name)))
                      ((functional-p leaf)
                       (let ((info (functional-info leaf)))
                         (if info
@@ -216,7 +254,7 @@
                             lvar-type)))
                      (t
                       lvar-type))))
-    (values type fun-name leaf)))
+    (values type fun-name leaf asserted)))
 
 (defun callable-argument-lossage-kind (fun-name leaf soft hard)
   (if (or (not leaf)
@@ -231,8 +269,8 @@
       hard))
 
 (defun validate-test-and-test-not (combination)
-  (let* ((comination-name (lvar-fun-name (combination-fun combination) t))
-         (info (info :function :info comination-name)))
+  (let* ((combination-name (lvar-fun-name (combination-fun combination) t))
+         (info (info :function :info combination-name)))
     (when (and info
                (ir1-attributep (fun-info-attributes info) call))
       (let (test
@@ -246,7 +284,7 @@
                                     (eq key :test-not))
                            (setf test-not value)))
                        (combination-args combination)
-                       (info :function :type comination-name))
+                       (global-ftype combination-name))
         (when (and test
                    test-not
                    (eq (type-intersection null-type (lvar-type test))
@@ -254,7 +292,7 @@
                    (eq (type-intersection null-type (lvar-type test-not))
                        *empty-type*))
           (note-lossage "~s: can't specify both :TEST and :TEST-NOT"
-                        comination-name))))))
+                        combination-name))))))
 
 (defun function-designator-lvar-types (annotation)
   (labels ((arg-type (arg)
@@ -277,9 +315,16 @@
       (labels ((%process-arg (spec)
                  (destructuring-bind (nth-arg . options) spec
                    (let* ((arg (nth nth-arg deps))
+                          (value-nth (getf options :value))
                           (key-nth (getf options :key))
+                          (value (and value-nth
+                                      (nth value-nth deps)))
                           (key (and key-nth (nth key-nth deps)))
-                          (key-return-type (cond ((not key)
+                          (key-return-type (cond (value-nth
+                                                  (if (lvar-p value)
+                                                      (lvar-type value)
+                                                      *universal-type*))
+                                                 ((not key)
                                                   nil)
                                                  ((lvar-p key)
                                                   (multiple-value-bind (type name) (lvar-fun-type key)
@@ -349,138 +394,170 @@
     (nreverse result)))
 
 (defun report-arg-count-mismatch (callee caller type arg-count
-                                  condition-type)
-  (multiple-value-bind (min max optional) (fun-type-arg-limits type)
-    (cond
-      ((and (not min) (not max))
-       nil)
-      ((not optional)
-       (when (/= arg-count min)
-         (warn condition-type
-               :format-control
-               "The function ~S is called by ~S with ~R argument~:P, but wants exactly ~R."
-               :format-arguments
-               (list
-                callee
-                caller
-                arg-count min))
-         t))
-      ((< arg-count min)
-       (warn condition-type
-             :format-control
-             "The function ~S is called by ~S with ~R argument~:P, but wants at least ~R."
-             :format-arguments
-             (list
-              callee
-              caller
-              arg-count min))
-       t)
-      ((not max)
-       nil)
-      ((> arg-count max)
-       (warn condition-type
-             :format-control
-             "The function ~S called by ~S with ~R argument~:P, but wants at most ~R."
-             :format-arguments
-             (list
-              callee
-              caller
-              arg-count max))
-       t))))
+                                  condition-type
+                                  &key lossage-fun)
+  (flet ((lose (format-control &rest format-args)
+           (if lossage-fun
+               (apply lossage-fun format-control format-args)
+               (warn condition-type :format-control format-control
+                                    :format-arguments format-args))
+           t))
+    (multiple-value-bind (min max optional) (fun-type-arg-limits type)
+      (or
+       (cond
+         ((and (not min) (not max))
+          nil)
+         ((not optional)
+          (when (/= arg-count min)
+            (lose
+             "The function ~S is called~@[ by ~S~] with ~R argument~:P, but wants exactly ~R."
+             callee caller
+             arg-count min)))
+         ((< arg-count min)
+          (lose
+           "The function ~S is called~@[ by ~S~] with ~R argument~:P, but wants at least ~R."
+           callee caller
+           arg-count min))
+         ((not max)
+          nil)
+         ((> arg-count max)
+          (lose
+           "The function ~S called~@[ by ~S~] with ~R argument~:P, but wants at most ~R."
+           callee caller
+           arg-count max)))
+       (let ((positional (fun-type-positional-count type)))
+         (when (and (fun-type-keyp type)
+                    (> arg-count positional)
+                    (oddp (- arg-count positional)))
+           (lose
+            "The function ~s is called with odd number of keyword arguments."
+            callee)))))))
+
+(defun disable-arg-count-checking (leaf type arg-count)
+  (when (lambda-p leaf)
+    (multiple-value-bind (min max) (fun-type-arg-limits type)
+      (when (and min
+                 (if max
+                     (<= min arg-count max)
+                     (<= min arg-count)))
+        (setf (lambda-lexenv leaf)
+              (make-lexenv :default (lambda-lexenv leaf)
+                           :policy (augment-policy verify-arg-count 0
+                                                   (lexenv-policy (lambda-lexenv leaf)))))))))
 
 ;;; This can provide better errors and better handle OR types than a
 ;;; simple type intersection.
 (defun check-function-designator-lvar (lvar annotation)
   (multiple-value-bind (type name leaf) (lvar-fun-type lvar)
-    (when (fun-type-p type)
-      ;; If the destination is a combination-fun that means the function
-      ;; is called here and not passed somewhere else, there's no longer a
-      ;; need to check the function type, the arguments to the call will
-      ;; do the same job.
-      (unless (let* ((dest (and lvar
-                                (lvar-dest lvar))))
-                (and (basic-combination-p dest)
-                     (eq (basic-combination-fun dest) lvar)))
-        (multiple-value-bind (args results)
-            (function-designator-lvar-types annotation)
-          (let* ((condition (callable-argument-lossage-kind name
-                                                            leaf
-                                                            'simple-style-warning
-                                                            'simple-warning))
-                 (type-condition (case condition
-                                   (simple-style-warning
-                                    'type-style-warning)
-                                   (t
-                                    'type-warning)))
-                 (caller (lvar-function-designator-annotation-caller annotation))
-                 (arg-count (length args)))
-            (or (report-arg-count-mismatch name caller
-                                           type
-                                           arg-count
-                                           condition)
-                (let ((param-types (fun-type-n-arg-types arg-count type)))
-                  (block nil
-                    ;; Need to check each OR seperately, a UNION could
-                    ;; intersect with the function parameters
-                    (labels ((hide-ors (current-or or-part)
-                               (loop for spec in args
-                                     collect (cond ((eq spec current-or)
-                                                    or-part)
-                                                   ((typep spec '(cons (eql or)))
-                                                    (sb!kernel::%type-union (cdr spec)))
-                                                   (t
-                                                    spec))))
-                             (check (arg param &optional
-                                                 current-spec)
-                               (when (eq (type-intersection param arg) *empty-type*)
-                                 (warn type-condition
-                                       :format-control
-                                       "The function ~S is called by ~S with ~S but it accepts ~S."
-                                       :format-arguments
-                                       (list
-                                        name
-                                        caller
-                                        (mapcar #'type-specifier (hide-ors current-spec arg))
-                                        (mapcar #'type-specifier param-types)))
-                                 (return t))))
-                      (loop for arg-type in args
-                            for param-type in param-types
-                            if (typep arg-type '(cons (eql or)))
-                            do (loop for type in (cdr arg-type)
-                                     do (check type param-type arg-type))
-                            else do (check arg-type param-type)))))
-                (let ((returns (single-value-type (fun-type-returns type))))
-                  (when (and (neq returns *wild-type*)
-                             (neq returns *empty-type*)
-                             (neq results *wild-type*)
-                             (eq (type-intersection returns results) *empty-type*))
-                    (warn type-condition
-                          :format-control
-                          "The function ~S called by ~S returns ~S but ~S is expected"
-                          :format-arguments
-                          (list
-                           name
-                           caller
-                           (type-specifier returns)
-                           (type-specifier results)))))))))
-      t)))
+    (cond
+      ((and name
+            (valid-function-name-p name)
+            (memq (info :function :kind name) '(:macro :special-form)))
+       (compiler-warn "~(~a~) ~s where a function is expected"
+                      (info :function :kind name) name))
+      ((fun-type-p type)
+       ;; If the destination is a combination-fun that means the function
+       ;; is called here and not passed somewhere else, there's no longer a
+       ;; need to check the function type, the arguments to the call will
+       ;; do the same job.
+       (unless (let* ((dest (lvar-dest lvar)))
+                 (and (basic-combination-p dest)
+                      (eq (basic-combination-fun dest) lvar)))
+         (multiple-value-bind (args results)
+             (function-designator-lvar-types annotation)
+           (let* ((condition (callable-argument-lossage-kind name
+                                                             leaf
+                                                             'simple-style-warning
+                                                             'simple-warning))
+                  (type-condition (case condition
+                                    (simple-style-warning
+                                     'type-style-warning)
+                                    (t
+                                     'type-warning)))
+                  (caller (lvar-function-designator-annotation-caller annotation))
+                  (arg-count (length args)))
+             (or (report-arg-count-mismatch name caller
+                                            type
+                                            arg-count
+                                            condition)
+                 (let ((param-types (fun-type-n-arg-types arg-count type)))
+                   (unless (and (eq caller 'reduce)
+                                (eql arg-count 2))
+                     (disable-arg-count-checking leaf type arg-count))
+                   (block nil
+                     ;; Need to check each OR seperately, a UNION could
+                     ;; intersect with the function parameters
+                     (labels ((hide-ors (current-or or-part)
+                                (loop for spec in args
+                                      collect (cond ((eq spec current-or)
+                                                     or-part)
+                                                    ((typep spec '(cons (eql or)))
+                                                     (sb-kernel::%type-union (cdr spec)))
+                                                    (t
+                                                     spec))))
+                              (check (arg param &optional
+                                                  current-spec)
+                                (when (eq (type-intersection param arg) *empty-type*)
+                                  (warn type-condition
+                                        :format-control
+                                        "The function ~S is called by ~S with ~S but it accepts ~S."
+                                        :format-arguments
+                                        (list
+                                         name
+                                         caller
+                                         (mapcar #'type-specifier (hide-ors current-spec arg))
+                                         (mapcar #'type-specifier param-types)))
+                                  (return t))))
+                       (loop for arg-type in args
+                             for param-type in param-types
+                             if (typep arg-type '(cons (eql or)))
+                             do (loop for type in (cdr arg-type)
+                                      do (check type param-type arg-type))
+                             else do (check arg-type param-type)))))
+                 (let ((returns (single-value-type (fun-type-returns type))))
+                   (when (and (neq returns *wild-type*)
+                              (neq returns *empty-type*)
+                              (neq results *wild-type*)
+                              (eq (type-intersection returns results) *empty-type*))
+                     (warn type-condition
+                           :format-control
+                           "The function ~S called by ~S returns ~S but ~S is expected"
+                           :format-arguments
+                           (list
+                            name
+                            caller
+                            (type-specifier returns)
+                            (type-specifier results)))))))))
+       t))))
 
 (defun check-function-lvar (lvar annotation)
   (let ((atype (lvar-function-annotation-type annotation)))
     (multiple-value-bind (type name leaf) (lvar-fun-type lvar)
       (when (fun-type-p type)
-        (let ((int (type-intersection type atype)))
-          (when (or (memq *empty-type* (fun-type-required int))
-                    (and (eq (fun-type-returns int) *empty-type*)
-                         (neq (fun-type-returns type) *empty-type*)
-                         (not (and (eq (fun-type-returns atype) *empty-type*)
-                                   (eq (fun-type-returns type) *wild-type*)))))
-            (%compile-time-type-error-warn annotation
-                                           (type-specifier atype)
-                                           (type-specifier type)
-                                           (list name)
-                                           :condition
-                                           (callable-argument-lossage-kind name
-                                                                           leaf
-                                                                           'type-style-warning
-                                                                           'type-warning))))))))
+        (let ((condition (callable-argument-lossage-kind name
+                                                         leaf
+                                                         'type-style-warning
+                                                         'type-warning)))
+          (if (eq (lvar-function-annotation-context annotation) :mv-call)
+              (let* ((*compiler-error-context* annotation)
+                     (max-accepted (nth-value 1 (fun-type-nargs (lvar-fun-type lvar))))
+                     (min-args (fun-type-nargs atype)))
+                (when (and max-accepted
+                           (> min-args max-accepted))
+                  (warn condition
+                        :format-control
+                        "~@<MULTIPLE-VALUE-CALL calls ~a with with at least ~R ~
+                              values when it expects at most ~R.~@:>"
+                        :format-arguments (list name min-args
+                                                max-accepted))))
+              (let ((int (type-intersection type atype)))
+                (when (or (memq *empty-type* (fun-type-required int))
+                          (and (eq (fun-type-returns int) *empty-type*)
+                               (neq (fun-type-returns type) *empty-type*)
+                               (not (and (eq (fun-type-returns atype) *empty-type*)
+                                         (eq (fun-type-returns type) *wild-type*)))))
+                  (%compile-time-type-error-warn annotation
+                                                 (type-specifier atype)
+                                                 (type-specifier type)
+                                                 (list name)
+                                                 :condition condition)))))))))
