@@ -560,6 +560,9 @@
 (defknown subseq (proper-sequence index &optional sequence-end) consed-sequence
   (flushable))
 
+(defknown vector-subseq* (vector index sequence-end) (simple-array * (*))
+  (flushable))
+
 (defknown copy-seq (proper-sequence) consed-sequence (flushable)
   :derive-type (sequence-result-nth-arg 0 :preserve-dimensions t))
 
@@ -1891,6 +1894,7 @@
   :derive-type #'result-type-first-arg)
 
 (defknown %cleanup-point (&rest t) t (reoptimize-when-unlinking))
+(defknown %dynamic-extent-start (t) t)
 (defknown %special-bind (t t) t)
 (defknown %special-unbind (&rest symbol) t)
 (defknown %listify-rest-args (t index) list (flushable))
@@ -1918,7 +1922,6 @@
 (defknown %%primitive (t &rest t) *)
 (defknown %pop-values (t) t)
 (defknown %nip-values (t t &rest t) (values))
-(defknown %dummy-dx-alloc (t t) t)
 (defknown %type-check-error (t t t) nil)
 (defknown %type-check-error/c (t t t) nil)
 
