@@ -14,15 +14,12 @@
 ;;; We need to define these predicates, since the TYPEP source
 ;;; transform picks whichever predicate was defined last when there
 ;;; are multiple predicates for equivalent types.
-(define-source-transform short-float-p (x) `(single-float-p ,x))
 #-long-float
 (define-source-transform long-float-p (x) `(double-float-p ,x))
 
 (define-source-transform compiled-function-p (x)
   (once-only ((x x))
-    `(and (functionp ,x)
-          #+(or sb-fasteval sb-eval)
-          (not (typep ,x 'interpreted-function)))))
+    `(and (functionp ,x) (not (funcallable-instance-p ,x)))))
 
 (define-source-transform char-int (x)
   `(char-code ,x))
@@ -33,14 +30,8 @@
 (deftransform make-symbol ((string) (simple-string))
   `(%make-symbol 0 string))
 
-#-immobile-space
-(define-source-transform %make-symbol (kind string)
-  (declare (ignore kind))
-  ;; Set "logically read-only" bit in pname.
-  `(sb-vm::%%make-symbol (logior-header-bits ,string ,sb-vm:+vector-shareable+)))
-
 ;;; We don't want to clutter the bignum code.
-#+(or x86 x86-64)
+#+(and (or x86 x86-64) (not bignum-assertions))
 (define-source-transform sb-bignum:%bignum-ref (bignum index)
   ;; KLUDGE: We use TRULY-THE here because even though the bignum code
   ;; is (currently) compiled with (SAFETY 0), the compiler insists on
@@ -93,10 +84,18 @@
                          sb-vm:bignum-digits-offset
                          index offset))
 
-(defun dd-contains-raw-slots-p (dd)
-  (dolist (dsd (dd-slots dd))
-    (unless (eq (dsd-raw-type dsd) t) (return t))))
-
+;;; When copying a structure, try to make the best decision possible
+;;; as to placement. This matters in a few circumstances:
+;;; - when the "system" mixed TLAB is different from the "user" mixed TLAB.
+;;; - when we distinguish boxed from mixed allocation to generation 0.
+;;; GIVE-UP-IR1-TRANSFORM might put the allocation in the wrong TLAB,
+;;; as the general fallback doesn't make the same distinctions.
+;;; Unfortunately the DEFSTRUCT-defined copiers were not inlining even when
+;;; explicitly requested, because COPY-SOMESTRUCT always transforms into
+;;; COPY-STRUCTURE, so we've lost any declared inline-ness of COPY-SOMESTRUCT.
+;;; Therefore we just try to infer it based on whether this transform is
+;;; "acting as" COPY-SOMESTRUCT for a particular struct whose copier
+;;; was requested to be inline.
 (deftransform copy-structure ((instance) * * :result result :node node)
   (let* ((classoid (lvar-type instance))
          (name (and (structure-classoid-p classoid) (classoid-name classoid)))
@@ -108,19 +107,22 @@
          (class-eq (and name
                         (eq (classoid-state classoid) :sealed)
                         (not (classoid-subclasses classoid))))
-         (dd (and class-eq (wrapper-info layout)))
+         (dd (and class-eq (layout-info layout)))
+         (dd-copier (and dd (sb-kernel::dd-copier-name dd)))
          (max-inlined-words 5))
     (unless (and result ; could be unused result (but entire call wasn't flushed?)
                  layout
-                 ;; And don't copy if raw slots are present on the precisely GCed backends.
-                 ;; To enable that, we'd want variants for {"all-raw", "all-boxed", "mixed"}
+                 ;; Fail if raw slots are present on the precisely GCed backends.
                  ;; Also note that VAR-ALLOC can not cope with dynamic-extent except where
                  ;; support has been added (x86oid so far); so requiring an exact type here
                  ;; causes VAR-ALLOC to become FIXED-ALLOC which works on more architectures.
-                 #-c-stack-is-control-stack (and dd (not (dd-contains-raw-slots-p dd)))
+                 #-c-stack-is-control-stack (and dd (not (dd-has-raw-slot-p dd)))
 
                  ;; Definitely do this if copying to stack
+                 ;; (Allocation has to be inlined, otherwise there's no way to DX it)
                  (or (lvar-dynamic-extent result)
+                     (and dd-copier
+                          (eq (sb-int:info :function :inlinep dd-copier) 'inline))
                      ;; Or if it's a small fixed number of words
                      ;; and speed at least as important as size.
                      (and class-eq
@@ -132,24 +134,27 @@
     ;;   depending on whether layouts are in immobile space.
     ;; - for a small number of slots, copying them is inlined
     (cond ((not dd) ; it's going to be some subtype of NAME
-           `(%copy-instance (%make-instance (%instance-length instance)) instance))
-          ((<= (dd-length dd) max-inlined-words)
-           `(let ((copy (%make-structure-instance ,dd nil)))
-              ;; ASSUMPTION: either %INSTANCE-REF is the correct accessor for this word,
-              ;; or the GC will treat random bit patterns as conservative pointers
-              ;; (i.e. not alter them if %INSTANCE-REF is not the correct accessor)
-              ,@(loop for i from sb-vm:instance-data-start below (dd-length dd)
-                      collect `(%instance-set copy ,i (%instance-ref instance ,i)))
-              copy))
-          (t
-           `(%copy-instance-slots (%make-structure-instance ,dd nil) instance)))))
+           ;; pessimistically assume MIXED rather than choosing at runtime
+           `(%copy-instance (%make-instance/mixed (%instance-length instance)) instance))
+          ((not (logtest (dd-flags dd) +dd-varylen+)) ; fixed length
+           ;; ASSUMPTION: either %INSTANCE-REF is the correct accessor for this word,
+           ;; or the GC will treat random bit patterns as conservative pointers
+           ;; (i.e. not alter them if %INSTANCE-REF is not the correct accessor)
+           (if (> (dd-length dd) max-inlined-words)
+               `(%copy-instance-slots (%make-structure-instance ,dd nil) instance)
+               `(let ((copy (%make-structure-instance ,dd nil)))
+                  ,@(loop for i from sb-vm:instance-data-start below (dd-length dd)
+                       collect `(%instance-set copy ,i (%instance-ref instance ,i)))
+                  copy)))
+          (t ; variable-length
+           `(let ((copy (,(if (dd-has-raw-slot-p dd) '%make-instance/mixed '%make-instance)
+                          (%instance-length instance))))
+              (%set-instance-layout copy (%instance-layout instance))
+              (%copy-instance-slots copy instance))))))
 
 (defun varying-length-struct-p (classoid)
-  ;; This is a nice feature to have in general, but at present it is only possible
-  ;; to make varying length instances of SB-VM:LAYOUT (or WRAPPER if that is the same type),
-  ;; and nothing else.
-  (eq classoid (load-time-value (find-classoid #+metaspace 'sb-vm:layout
-                                               #-metaspace 'wrapper))))
+  (let ((dd (find-defstruct-description (classoid-name classoid))))
+    (logtest (dd-flags dd) +dd-varylen+)))
 
 (deftransform %instance-length ((instance))
   (let ((classoid (lvar-type instance)))
@@ -159,11 +164,17 @@
              (not (varying-length-struct-p classoid))
              ;; TODO: if sealed with subclasses which add no slots, use the fixed length
              (not (classoid-subclasses classoid)))
-        (dd-length (wrapper-dd (sb-kernel::compiler-layout-or-lose (classoid-name classoid))))
+        (dd-length (layout-dd (sb-kernel::compiler-layout-or-lose (classoid-name classoid))))
         (give-up-ir1-transform))))
 
-(define-source-transform %instance-wrapper (x) `(layout-friend (%instance-layout ,x)))
-(define-source-transform %fun-wrapper (x) `(layout-friend (%fun-layout ,x)))
+;;; This doesn't help a whole lot, but it does fire during compilation of 'info-vector'
+;;; which uses variable-length instances of PACKED-INFO, having no slot transforms.
+(define-source-transform (setf %instance-ref) (newval instance index)
+  `(let ((.newval. ,newval)
+         (.instance. ,instance)
+         (.index. ,index))
+     (%instance-set .instance. .index. .newval.)
+     .newval.))
 
 ;;; *** These transforms should be the only code, aside from the C runtime
 ;;;     with knowledge of the layout index.
@@ -172,9 +183,9 @@
 #-compact-instance-header
 (progn
   (define-source-transform %instance-layout (x)
-    `(truly-the sb-vm:layout (%instance-ref ,x 0)))
+    `(truly-the layout (%instance-ref ,x 0)))
   (define-source-transform %set-instance-layout (instance layout)
-    `(%instance-set ,instance 0 (the sb-vm:layout ,layout)))
+    `(%instance-set ,instance 0 (the layout ,layout)))
   (define-source-transform function-with-layout-p (x)
     `(funcallable-instance-p ,x)))
 
@@ -276,12 +287,12 @@
         ;; so explicitly return the NEW-VALUE
         `(typecase string
            ((simple-array character (*))
-            (let ((c (the* (character :context :aref) new-value)))
+            (let ((c (the* (character :context 'aref-context) new-value)))
               (data-vector-set string index c)
               c))
            #+sb-unicode
            ((simple-array base-char (*))
-            (let ((c (the* (base-char :context :aref :silent-conflict t) new-value)))
+            (let ((c (the* (base-char :context 'aref-context :silent-conflict t) new-value)))
               (data-vector-set string index c)
               c))))))
 
@@ -531,48 +542,6 @@
           (if (zerop (lvar-value item)) '(- length count) 'count)
           '(if (zerop item) (- length count) count))))
 
-;;; This transform does not require that ITEM be derived as BIT,
-;;; but at runtime it has to be.
-(deftransform fill ((sequence item) (simple-bit-vector t) *
-                    :policy (>= speed space))
-  `(let ((value (logand (- (the bit item)) most-positive-word)))
-     ;; Unlike for the SIMPLE-BASE-STRING case, we are allowed to touch
-     ;; bits beyond LENGTH with impunity.
-     (dotimes (index (ceiling (vector-length sequence) sb-vm:n-word-bits))
-       (declare (optimize (speed 3) (safety 0))
-                (type index index))
-       (setf (%vector-raw-bits sequence index) value))
-     sequence))
-
-(deftransform fill ((sequence item) (simple-base-string t) *
-                                    :policy (>= speed space))
-  (let ((multiplier (logand #x0101010101010101 most-positive-word)))
-    `(let* ((value ,(if (and (constant-lvar-p item) (typep item 'base-char))
-                        (* multiplier (char-code (lvar-value item)))
-                        ;; Use multiplication if it's known to be cheap
-                        #+(or x86 x86-64)
-                        `(* ,multiplier (char-code (the base-char item)))
-                        #-(or x86 x86-64)
-                        '(let ((code (char-code (the base-char item))))
-                          (setf code (dpb code (byte 8 8) code))
-                          (setf code (dpb code (byte 16 16) code))
-                          #+64-bit (dpb code (byte 32 32) code))))
-            (len (vector-length sequence))
-            (words (truncate len sb-vm:n-word-bytes)))
-       (dotimes (index words)
-         (declare (optimize (speed 3) (safety 0))
-                  (type index index))
-         (setf (%vector-raw-bits sequence index) value))
-       ;; For 64-bit:
-       ;;  if 1 more byte should be written, then shift-towards-start 56
-       ;;  if 2 more bytes ...               then shift-towards-start 48
-       ;;  etc
-       ;; This correctly rewrites the trailing null in its proper place.
-       (let ((bits (ash (mod len sb-vm:n-word-bytes) 3)))
-         (when (plusp bits)
-           (setf (%vector-raw-bits sequence words)
-                 (shift-towards-start value (- bits)))))
-       sequence)))
 
 ;;;; %BYTE-BLT
 
@@ -653,7 +622,7 @@
                      (push `(define-good-modular-fun ,fun :untagged t) result)
                      (push `(define-good-modular-fun ,fun :tagged t) result))))))
   (define-good-signed-modular-funs
-      logand logandc1 logandc2 logeqv logior lognand lognor lognot
+      logand logandc2 logeqv logior lognand lognor lognot
       logorc1 logorc2 logxor))
 
 ;;;; word-wise logical operations
@@ -689,7 +658,7 @@
   `(logand (logorc2 x y) ,most-positive-word))
 
 (deftransform word-logical-andc1 ((x y))
-  `(logand (logandc1 x y) ,most-positive-word))
+  `(logand (logandc2 y x) ,most-positive-word))
 
 (deftransform word-logical-andc2 ((x y))
   `(logand (logandc2 x y) ,most-positive-word))
@@ -757,14 +726,29 @@
   (logior sb-vm:character-widetag
           (ash (char-code (lvar-value obj)) sb-vm:n-widetag-bits)))
 
+;;; FIXME: The following should really be done by defining
+;;; UNBOUND-MARKER as a primitive object.
 ;; So that the PCL code walker doesn't observe any use of %PRIMITIVE,
 ;; MAKE-UNBOUND-MARKER is an ordinary function, not a macro.
 #-sb-xc-host
 (defun make-unbound-marker () ; for interpreters
   (sb-sys:%primitive make-unbound-marker))
-;; Get the main compiler to transform MAKE-UNBOUND-MARKER
-;; without the fopcompiler seeing it - the fopcompiler does
-;; expand compiler-macros, but not source-transforms -
-;; because %PRIMITIVE is not generally fopcompilable.
+;; Get the main compiler to transform MAKE-UNBOUND-MARKER.
 (sb-c:define-source-transform make-unbound-marker ()
   `(sb-sys:%primitive make-unbound-marker))
+
+(deftransform (cas symbol-value) ((old new symbol))
+  (let ((cname (and (constant-lvar-p symbol) (lvar-value symbol))))
+    (case (and cname (info :variable :kind cname))
+      ((:special :global)
+       (let ((type (info :variable :type cname)))
+         `(truly-the ,type
+                     (%compare-and-swap-symbol-value ',cname old (the ,type new)))))
+      (t
+       `(progn
+          (about-to-modify-symbol-value symbol 'compare-and-swap new)
+          (%compare-and-swap-symbol-value symbol old new))))))
+
+(deftransform (cas svref) ((old new vector index))
+  '(let ((v (the simple-vector vector)))
+    (%compare-and-swap-svref v (check-bound v (length v) index) old new)))
