@@ -100,6 +100,14 @@
   (declare (type lvar lvar))
   (ir2-lvar-primitive-type (lvar-info lvar)))
 
+;;; Return true if a constant LEAF is of a type which we can legally
+;;; directly reference in code. Named constants with arbitrary pointer
+;;; values cannot, since we must preserve EQLness.
+(defun legal-immediate-constant-p (leaf)
+  (declare (type constant leaf))
+  (or (not (leaf-has-source-name-p leaf))
+      (sb-xc:typep (constant-value leaf) '(or symbol number character))))
+
 ;;; If LVAR is used only by a REF to a leaf that can be delayed, then
 ;;; return the leaf, otherwise return NIL.
 (defun lvar-delayed-leaf (lvar)
@@ -127,7 +135,9 @@
      (t (let ((tn (make-normal-tn (ir2-lvar-primitive-type info))))
           (setf (tn-type tn) (lvar-type lvar))
           (setf (ir2-lvar-locs info) (list tn))
-          (when (lvar-dynamic-extent lvar)
+          (when (and (lvar-dynamic-extent lvar)
+                     (eq (dx-info-value (lvar-dynamic-extent lvar))
+                         lvar))
             (setf (ir2-lvar-stack-pointer info)
                   (make-stack-pointer-tn)))))))
   (ltn-annotate-casts lvar)
@@ -162,11 +172,10 @@
   (values))
 
 #+call-symbol
-(defoptimizer (%coerce-callable-for-call ltn-annotate) ((fun) node ltn-policy)
-  (declare (ignore ltn-policy))
+(defoptimizer (%coerce-callable-for-call ltn-annotate) ((fun) node)
   (multiple-value-bind (dest dest-lvar)
       (and (node-lvar node)
-           (principal-lvar-dest-and-lvar (node-lvar node)))
+           (principal-lvar-end (node-lvar node)))
     (cond ((and (basic-combination-p dest)
                 (eq (basic-combination-kind dest) :full)
                 (eq (basic-combination-fun dest) dest-lvar)
@@ -174,7 +183,8 @@
                 ;; bother optimizing it.
                 (not (lvar-value-is-nil fun)))
            (setf (basic-combination-fun dest) fun
-                 (basic-combination-args node) '(nil)
+                 (basic-combination-args node) '()
+                 (lvar-dest fun) dest
                  (node-lvar node) nil
                  (lvar-info fun) (make-ir2-lvar (primitive-type (lvar-type fun))))
            (annotate-1-value-lvar fun))
@@ -190,9 +200,15 @@
 (defun flush-full-call-tail-transfer (call)
   (declare (type basic-combination call))
   (let ((tails (and (node-tail-p call)
-                    (lambda-tail-set (node-home-lambda call)))))
+                    (lambda-tail-set (node-home-lambda call))))
+        (unboxed-return (let ((info (basic-combination-fun-info call)))
+                          (and info
+                               (ir1-attributep (fun-info-attributes info) unboxed-return)))))
     (cond ((not tails))
-          ((eq (return-info-kind (tail-set-info tails)) :unknown)
+          ((eq (return-info-kind (tail-set-info tails))
+               (if unboxed-return
+                   :unboxed
+                   :unknown))
            (ir2-change-node-successor call
                                       (component-tail (block-component (node-block call)))))
           (t
@@ -289,8 +305,9 @@
     (setf (lvar-info lvar) info)
     (when (lvar-dynamic-extent lvar)
       (aver (proper-list-of-length-p types 1))
-      (setf (ir2-lvar-stack-pointer info)
-            (make-stack-pointer-tn))))
+      (when (eq (dx-info-value (lvar-dynamic-extent lvar)) lvar)
+        (setf (ir2-lvar-stack-pointer info)
+              (make-stack-pointer-tn)))))
   (ltn-annotate-casts lvar)
   (values))
 
@@ -490,8 +507,9 @@
       (let ((info (make-ir2-lvar *backend-t-primitive-type*)))
         (setf (lvar-info lvar) info)
         (setf (ir2-lvar-kind info) :delayed)
-        (setf (ir2-lvar-stack-pointer info)
-              (make-stack-pointer-tn))))))
+        (when (eq (dx-info-value (lvar-dynamic-extent lvar)) lvar)
+          (setf (ir2-lvar-stack-pointer info)
+                (make-stack-pointer-tn)))))))
 
 ;;; We need a special method for %UNWIND-PROTECT that ignores the
 ;;; cleanup function. We don't annotate either arg, since we don't
@@ -500,24 +518,16 @@
 ;;; (The default is o.k. for %CATCH, since environment analysis
 ;;; converted the reference to the escape function into a constant
 ;;; reference to the NLX-INFO.)
-(defoptimizer (%unwind-protect ltn-annotate) ((escape cleanup)
-                                              node
-                                              ltn-policy)
-  (declare (ignore escape cleanup ltn-policy))
+(defoptimizer (%unwind-protect ltn-annotate) ((escape cleanup) node)
   (setf (basic-combination-info node) :funny)
   (setf (node-tail-p node) nil))
 
 ;;; Make sure that arguments of magic functions are not annotated.
 ;;; (Otherwise the compiler may dump its internal structures as
 ;;; constants :-()
-(defoptimizer (%pop-values ltn-annotate) ((%lvar) node ltn-policy)
-  (declare (ignore %lvar node ltn-policy)))
-(defoptimizer (%dummy-dx-alloc ltn-annotate) ((target source) node ltn-policy)
-  (declare (ignore target source node ltn-policy)))
+(defoptimizer (%pop-values ltn-annotate) ((%lvar)))
 
-(defoptimizer (%nip-values ltn-annotate) ((&rest lvars)
-                                          node ltn-policy)
-  (declare (ignore node ltn-policy))
+(defoptimizer (%nip-values ltn-annotate) ((&rest lvars))
   ;; Undo the optimization performed by LTN-ANALYZE-MV-CALL,
   ;; which only uses the CSP of the first argument.
   (loop for lvar-lvar in lvars
@@ -556,7 +566,13 @@
                       (funcall (second type) value)
                       (sb-xc:typep value type))))
           (cond (lvar
-                 (and (constant-lvar-p lvar)
+                 (and (if (policy (let ((uses (lvar-uses lvar)))
+                                    (if (consp uses)
+                                        (car uses)
+                                        uses))
+                              (= preserve-constants 3))
+                          (constant-lvar-ignore-types-p lvar nil)
+                          (constant-lvar-p lvar))
                       (type-p (lvar-value lvar) (cdr restr))))
                 (tn
                  (and (eq (tn-kind tn) :constant)
@@ -617,7 +633,7 @@
 ;;; before we run out of restrictions, then we only succeed if the
 ;;; leftover restrictions are *. If we run out of restrictions before
 ;;; we run out of result types, then we always win.
-(defun template-results-ok (template result-type)
+(defun template-results-ok (template result-type &optional single-value-p)
   (declare (type template template)
            (type ctype result-type))
   (when (template-more-results-type template)
@@ -625,19 +641,28 @@
   (let ((types (template-result-types template)))
     (cond
      ((values-type-p result-type)
-      (do ((ltypes (append (args-type-required result-type)
-                           (args-type-optional result-type))
-                   (rest ltypes))
-           (types types (rest types)))
-          ((null ltypes)
-           (dolist (type types t)
-             (unless (eq type '*)
-               (return nil))))
-        (when (null types) (return t))
-        (let ((type (first types)))
-          (unless (operand-restriction-ok type
-                                          (primitive-type (first ltypes)))
-            (return nil)))))
+      (if (and single-value-p
+               (let ((optional (vop-info-optional-results template)))
+                 (and optional
+                      (not (eql (car optional) 0))
+                      (= (length optional)
+                         (1- (length types))))))
+          (operand-restriction-ok (car types)
+                                  (primitive-type (or (car (args-type-required result-type))
+                                                      (car (args-type-optional result-type)))))
+          (do ((ltypes (append (args-type-required result-type)
+                               (args-type-optional result-type))
+                       (rest ltypes))
+               (types types (rest types)))
+              ((null ltypes)
+               (dolist (type types t)
+                 (unless (eq type '*)
+                   (return nil))))
+            (when (null types) (return t))
+            (let ((type (first types)))
+              (unless (operand-restriction-ok type
+                                              (primitive-type (first ltypes)))
+                (return nil))))))
      (types
       (operand-restriction-ok (first types) (primitive-type result-type)))
      (t t))))
@@ -677,7 +702,7 @@
                           (immediately-used-p (if-test dest) call))
                      (values t nil)
                      (values nil :conditional)))))
-          ((template-results-ok template dtype)
+          ((template-results-ok template dtype (lvar-single-value-p (node-lvar call)))
            (values t nil))
           (t
            (values nil :result-types)))))
