@@ -194,8 +194,8 @@
   ;; the optimizer for this node type doesn't care, it can elect not
   ;; to clear this flag.
   (reoptimize t :type boolean)
-  ;; if the LVAR value is DYNAMIC-EXTENT, CLEANUP protecting it.
-  (dynamic-extent nil :type (or null cleanup))
+  ;; if the LVAR value is DYNAMIC-EXTENT, some information.
+  (dynamic-extent nil :type (or null dx-info))
   ;; something or other that the back end annotates this lvar with
   (info nil)
   ;; Nodes to reoptimize together with the lvar
@@ -204,6 +204,25 @@
   (dependent-annotations nil))
 (!set-load-form-method lvar (:xc :target) :ignore-it)
 
+;;; A DX-INFO structure is used to accumulate information about a
+;;; dynamic extent declaration.
+(defstruct (dx-info (:copier nil))
+  ;; The kind of dynamic extent this is.
+  (kind (missing-arg) :type (member enclose dynamic-extent truly-dynamic-extent))
+  ;; The value recognized to be declared dynamic extent.
+  (value (missing-arg) :type lvar)
+  ;; The stack-allocatable values in the transitive closure of the
+  ;; relation determined by the "otherwise inaccessible part"
+  ;; criterion. This is filled in by environment analysis.
+  (subparts nil :type list)
+  ;; The CLEANUP associated with this dynamic extent.
+  (cleanup (missing-arg) :type cleanup))
+
+(defprinter (dx-info :identity t)
+  kind
+  value
+  subparts)
+
 ;;; These are used for annotating a LVAR with information that can't
 ;;; be expressed using types or if the CAST semantics are undesirable
 ;;; (type derivation, runtime errors).
@@ -348,7 +367,7 @@
 
 (defun %with-ir1-environment-from-node (node fun)
   (declare (type node node) (type function fun))
-  #-sb-xc-host (declare (dynamic-extent fun)) ; "unable"
+  (declare (dynamic-extent fun))
   (let ((*current-component* (node-component node))
         (*lexenv* (node-lexenv node))
         (*current-path* (node-source-path node)))
@@ -399,6 +418,7 @@
   (defattr block-type-check)
   (defattr block-delete-p))
 
+;;; The LOOP structure holds information about a loop.
 (defstruct (cloop (:conc-name loop-)
                   (:predicate loop-p)
                   (:constructor make-loop)
@@ -483,13 +503,14 @@
   ;; other sets used in constraint propagation and/or copy propagation
   (in nil)
   (out nil)
-  ;; Set of all blocks that dominate this block. NIL is interpreted
-  ;; as "all blocks in component".
-  (dominators nil :type (or null sset))
-  ;; the LOOP that this block belongs to
-  (loop nil :type (or null cloop))
-  ;; next block in the loop.
+  ;; The Loop structure for the innermost loop that contains this
+  ;; block. Null only temporarily.
+  (loop nil :type (or cloop null))
+  ;; A link that holds together the list of blocks within Loop. Null
+  ;; at the end or when we haven't determined it yet.
   (loop-next nil :type (or null cblock))
+  ;; The immediate dominator of this block.
+  (dominator nil :type (or null cblock))
   ;; the component this block is in, or NIL temporarily during IR1
   ;; conversion and in deleted blocks
   (component (progn
@@ -544,12 +565,7 @@
 ;;;   component.
 (defstruct (component (:copier nil)
                       (:constructor make-component
-                       (head
-                        tail &aux
-                        (last-block tail)
-                        (outer-loop (make-loop :kind :outer
-                                               :head head
-                                               :tail (list tail))))))
+                       (head tail &aux (last-block tail))))
   ;; space where this component will be allocated in
   ;; :auto won't make any codegen optimizations pertinent to immobile space,
   ;; but will place the code there given sufficient available space.
@@ -643,7 +659,7 @@
   ;; this is filled by environment analysis
   (dx-lvars nil :type list)
   ;; The default LOOP in the component.
-  (outer-loop (missing-arg) :type cloop)
+  (outer-loop (make-loop :kind :outer :head head :tail (list tail)) :type cloop)
   (max-block-number 0 :type fixnum)
   (dominators-computed nil))
 
@@ -695,13 +711,9 @@
   ;; structures whose NLX-INFO-CLEANUP is this cleanup. This is filled
   ;; in by environment analysis.
   ;;
-  ;; For :DYNAMIC-EXTENT: a list of all DX LVARs, preserved by this
-  ;; cleanup. This is filled when the cleanup is created (now by
-  ;; locall call analysis) and is rechecked by environment
-  ;; analysis. (For closures this is a list of the LVAR of the enclose
-  ;; after environment analysis.)
-  (nlx-info nil :type list)
-  (dx-kind nil))
+  ;; For :DYNAMIC-EXTENT: a list of all DX-INFOs, preserved by this
+  ;; cleanup.
+  (nlx-info nil :type list))
 (defprinter (cleanup :identity t)
   kind
   mess-up
@@ -769,8 +781,7 @@
   ;; this thus provides a good indication of what kind of exit is
   ;; being done.
   (cleanup (missing-arg) :type cleanup)
-  ;; the ``continuation'' exited to (the block, succeeding the EXIT
-  ;; nodes). If this exit is from an escape function (CATCH or
+  ;; the continuation exited to. If this exit is from an escape function (CATCH or
   ;; UNWIND-PROTECT), then environment analysis deletes the
   ;; escape function and instead has the %NLX-ENTRY use this
   ;; continuation.
@@ -855,17 +866,11 @@
   ;; This may be non-nil when REFS and SETS are null, since code can be deleted.
   (ever-used nil :type (member nil set t))
   ;; is it declared dynamic-extent, or truly-dynamic-extent?
-  (extent nil :type (member nil truly-dynamic-extent dynamic-extent indefinite-extent
-                            dynamic-extent-no-note))
+  (dynamic-extent nil :type (member nil truly-dynamic-extent dynamic-extent))
   ;; some kind of info used by the back end
   (info nil))
 (!set-load-form-method leaf (:xc :target) :ignore-it)
 
-(defun leaf-dynamic-extent (leaf)
-  (let ((extent (leaf-extent leaf)))
-    (unless (member extent '(nil indefinite-extent))
-      extent)))
-
 ;;; LEAF name operations
 (defun leaf-has-source-name-p (leaf)
   (not (eq (leaf-%source-name leaf)
@@ -1111,9 +1116,8 @@
   ;; xref information for this functional (only used for functions with an
   ;; XEP)
   (xref () :type list)
-  ;; True if this functional was created from an inline expansion. This
-  ;; is either T, or the GLOBAL-VAR for which it is an expansion.
-  (inline-expanded nil)
+  ;; True if this functional was created from an inline expansion.
+  (inline-expanded nil :type boolean)
   ;; Is it coming from a top-level NAMED-LAMBDA?
   (top-level-defun-p nil)
   (ignore nil))
@@ -1143,9 +1147,6 @@
       ;; it looks as though it's never interesting to get debug names
       ;; from them, so it's moot. -- WHN)
       (leaf-source-name leaf)))
-(defun leaf-%debug-name (leaf)
-  (when (functional-p leaf)
-    (functional-%debug-name leaf)))
 
 ;;; Is FUNCTIONAL LET-converted? (where we're indifferent to whether
 ;;; it returns one value or multiple values)
@@ -1183,7 +1184,6 @@
 (defstruct (clambda (:include functional)
                     (:conc-name lambda-)
                     (:predicate lambda-p)
-                    (:constructor make-lambda)
                     (:copier nil))
   ;; list of LAMBDA-VAR descriptors for arguments
   (vars nil :type list)
@@ -1218,15 +1218,8 @@
   ;; depends on in such a way that DFO shouldn't put them in separate
   ;; components.
   (calls-or-closes (make-sset) :type (or null sset))
-  ;; the TAIL-SET that this LAMBDA is in. This is null during creation.
-  ;;
-  ;; In CMU CL, and old SBCL, this was also NILed out when LET
-  ;; conversion happened. That caused some problems, so as of
-  ;; sbcl-0.pre7.37.flaky5.2 when I was trying to get the compiler to
-  ;; emit :EXTERNAL functions directly, and so now the value
-  ;; is no longer NILed out in LET conversion, but instead copied
-  ;; (so that any further optimizations on the rest of the tail
-  ;; set won't modify the value) if necessary.
+  ;; the TAIL-SET that this LAMBDA is in. This is null during creation
+  ;; and in let lambdas.
   (tail-set nil :type (or tail-set null))
   ;; the structure which represents the phsical environment that this
   ;; function's variables are allocated in. This is filled in by
@@ -1471,9 +1464,7 @@
                 (:copier nil))
   ;; The leaf referenced.
   (leaf nil :type leaf)
-  ;; CONSTANT nodes are always anonymous, since we wish to coalesce named and
-  ;; unnamed constants that are equivalent, we need to keep track of the
-  ;; reference name for XREF.
+  ;; KLUDGE: This is supposed to help with keyword debug messages somehow.
   (%source-name (missing-arg) :type symbol :read-only t)
   ;; Constraints that cannot be expressed as NODE-DERIVED-TYPE
   (constraints nil))
@@ -1628,10 +1619,10 @@
   ;; assertion is satisfied:
   ;;
   ;; NIL
-  ;;    No type check is necessary (VALUE type is a subtype of the TYPE-TO-CHECK.)
+  ;;    No type check is necessary (proven type of VALUE is a subtype of the TYPE-TO-CHECK.)
   ;;
   ;; :EXTERNAL
-  ;;    Type check will be performed by NODE-DEST.
+  ;;    A type check will be performed by NODE-DEST.
   ;;
   ;; T
   ;;    A type check is needed.
@@ -1720,8 +1711,11 @@
 ;;; would be emitted, if necessary.
 (defstruct (enclose (:include valued-node) ; this node uses a dummy lvar for dx analysis
                     (:copier nil))
-  ;; the list of functionals that this ENCLOSE node allocates.
-  (funs nil :type list))
+  ;; The list of functionals that this ENCLOSE node allocates.
+  (funs nil :type list)
+  ;; The cleanup for this enclose if any of its functionals are
+  ;; declared dynamic extent.
+  (cleanup nil :type (or null cleanup)))
 (defprinter (enclose :identity t)
   funs)
 
