@@ -547,8 +547,7 @@
   (:generator 6
     (check-ocfp-and-return-pc old-fp return-pc)
     ;; Zot all of the stack except for the old-fp and return-pc.
-    (inst mov rsp-tn rbp-tn)
-    (inst pop rbp-tn)
+    (inst leave)
     (inst ret)))
 
 ;;;; full call
@@ -614,7 +613,7 @@
                    `((args :more t ,@(unless (eq args :fixed)
                                        '(:scs (descriptor-reg control-stack)))))))
 
-               ,@(when (eq return :fixed)
+               ,@(when (memq return '(:fixed :unboxed))
                    '((:results (values :more t))))
 
                (:save-p ,(if (eq return :tail) :compute-only t))
@@ -639,8 +638,10 @@
                    '(fun-type)))
 
                (:ignore
-               ,@(unless (or variable (eq return :tail)) '(arg-locs))
-               ,@(unless variable '(args)))
+                ,@(unless (or variable (eq return :tail)) '(arg-locs))
+                ,@(unless variable '(args))
+                ,@(when (eq return :unboxed)
+                    '(values)))
 
                ;; We pass either the fdefn object (for named call) or
                ;; the actual function object (for unnamed call) in
@@ -731,7 +732,7 @@
                           ;; (without-interrupts ...)
                           ;;
                           ;; dtc; Could be doing a tail call from a
-                          ;; known-local-call etc in which the old-fp
+                          ;; known-call-local etc in which the old-fp
                           ;; or ret-pc are in regs or in non-standard
                           ;; places. If the passing location were
                           ;; wired to the stack in standard locations
@@ -818,7 +819,7 @@
                     '((note-this-location vop :unknown-return)
                       (receive-unknown-values values-start nvals start count
                                               node)))
-                   (:tail))))))
+                   ((:tail :unboxed)))))))
 
   (define-full-call call nil :fixed nil)
   (define-full-call call-named t :fixed nil)
@@ -836,7 +837,10 @@
   (define-full-call call-variable nil :fixed t)
   (define-full-call multiple-call-variable nil :unknown t)
   (define-full-call fixed-call-named t :fixed nil :fixed)
-  (define-full-call fixed-tail-call-named t :tail nil :fixed))
+  (define-full-call fixed-tail-call-named t :tail nil :fixed)
+
+  (define-full-call unboxed-call-named t :unboxed nil)
+  (define-full-call fixed-unboxed-call-named t :unboxed nil :fixed))
 
 ;;; Call NAME "directly" meaning in a single JMP or CALL instruction,
 ;;; if possible (without loading RAX)
@@ -944,12 +948,10 @@
   (:ignore value)
   (:generator 6
     (check-ocfp-and-return-pc old-fp return-pc)
-    ;; Drop stack above old-fp
-    (inst mov rsp-tn rbp-tn)
+    ;; Drop stack above old-fp and restore old frame pointer
+    (inst leave)
     ;; Clear the multiple-value return flag
     (inst clc)
-    ;; Restore the old frame pointer
-    (inst pop rbp-tn)
     ;; And return.
     (inst ret)))
 
@@ -1003,8 +1005,7 @@
     ;; stack and we've changed the stack pointer. So we have to
     ;; tell it to index off of RBX instead of RBP.
     (cond ((<= nvals register-arg-count)
-           (inst mov rsp-tn rbp-tn)
-           (inst pop rbp-tn)
+           (inst leave)
            (inst ret))
           (t
            ;; Some values are on the stack after RETURN-PC and OLD-FP,
@@ -1051,11 +1052,10 @@
         ;; Return with one value.
         (loadw a0 vals -1)
         ;; Clear the stack until ocfp.
-        (inst mov rsp-tn rbp-tn)
+        (inst leave)
         ;; clear the multiple-value return flag
         (inst clc)
         ;; Out of here.
-        (inst pop rbp-tn)
         (inst ret)
         ;; Nope, not the single case. Jump to the assembly routine.
         (emit-label not-single)))
@@ -1084,19 +1084,19 @@
   (:generator 20
     ;; Avoid the copy if there are no more args.
     (cond ((zerop fixed)
-           (inst test rcx-tn rcx-tn)
+           (inst test :dword rcx-tn rcx-tn)
            (inst jmp :z JUST-ALLOC-FRAME))
           ((and (eql min-verified fixed)
                 (> fixed 1))
            ;; verify-arg-count will do a CMP
            (inst jmp :e JUST-ALLOC-FRAME))
           (t
-           (inst cmp rcx-tn (fixnumize fixed))
+           (inst cmp :dword rcx-tn (fixnumize fixed))
            (inst jmp :be JUST-ALLOC-FRAME)))
 
     ;; Create a negated copy of the number of arguments to allow us to
     ;; use EA calculations in order to do scaled subtraction.
-    (inst mov temp rcx-tn)
+    (inst mov :dword temp rcx-tn)
     (inst neg temp)
 
     ;; Allocate the space on the stack.
@@ -1291,23 +1291,23 @@
   (:results (result :scs (descriptor-reg)))
   (:node-var node)
   (:generator 20
-    (let ((loop (gen-label))
-          (done (gen-label))
-          (leave-pa (gen-label))
-          (stack-allocate-p (node-stack-allocate-p node)))
-      ;; Compute the number of bytes to allocate
-      (let ((shift (- (1+ word-shift) n-fixnum-tag-bits)))
-        (if (location= count rcx)
-            (inst shl :dword rcx shift)
-            (inst lea :dword rcx (ea nil count (ash 1 shift)))))
-      ;; Setup for the CDR of the last cons (or the entire result) being NIL.
-      (inst mov result nil-value)
-      (inst jrcxz DONE)
-      (unless stack-allocate-p
-        (instrument-alloc +cons-primtype+ rcx node (list value dst) thread-tn))
-      (pseudo-atomic (:elide-if stack-allocate-p :thread-tn thread-tn)
+    ;; Compute the number of bytes to allocate
+    (let ((shift (- (1+ word-shift) n-fixnum-tag-bits)))
+      (if (location= count rcx)
+          (inst shl :dword rcx shift)
+          (inst lea :dword rcx (ea nil count (ash 1 shift)))))
+    ;; Setup for the CDR of the last cons (or the entire result) being NIL.
+    (inst mov result nil-value)
+    (cond ((not (member :allocation-size-histogram sb-xc:*features*))
+           (inst jrcxz DONE))
+          (t ; jumps too far for JRCXZ sometimes
+           (inst test rcx rcx)
+           (inst jmp :z done)))
+    (unless (node-stack-allocate-p node)
+      (instrument-alloc +cons-primtype+ rcx node (list value dst) thread-tn))
+    (pseudo-atomic (:elide-if (node-stack-allocate-p node) :thread-tn thread-tn)
        ;; Produce an untagged pointer into DST
-       (if stack-allocate-p
+       (if (node-stack-allocate-p node)
            (stack-allocation rcx 0 dst)
            (allocation +cons-primtype+ rcx 0 dst node value thread-tn
                        :overflow
@@ -1325,7 +1325,7 @@
        ;; The rightmost arguments are at lower addresses.
        ;; Start by indexing the last argument
        (inst neg rcx) ; :QWORD because it's a signed number
-       (emit-label LOOP)
+       LOOP
        ;; Grab one value and store into this cons. Use RCX as an index into the
        ;; vector of values in CONTEXT, but add 8 because CONTEXT points exactly at
        ;; the 0th value, which means that the index is 1 word too low.
@@ -1340,8 +1340,8 @@
        (inst sub dst (* cons-size n-word-bytes)) ; get the preceding cons
        (inst inc rcx) ; :QWORD because it's a signed number
        (inst jmp :nz loop)
-       (emit-label leave-pa))
-      (emit-label done))))
+       LEAVE-PA)
+    DONE))
 
 ;;; Return the location and size of the &MORE arg glob created by
 ;;; COPY-MORE-ARG. SUPPLIED is the total number of arguments supplied
@@ -1387,21 +1387,21 @@
             (generate-error-code vop 'invalid-arg-count-error nargs)))
       (cond ((not min)
              (if (zerop max)
-                 (inst test nargs nargs)
-                 (inst cmp nargs (fixnumize max)))
+                 (inst test :dword nargs nargs)
+                 (inst cmp :dword nargs (fixnumize max)))
              (inst jmp :ne err-lab))
             (max
              (if (zerop min)
                  (setf temp nargs)
-                 (inst lea temp (ea (fixnumize (- min)) nargs)))
-             (inst cmp temp (fixnumize (- max min)))
+                 (inst lea :dword temp (ea (fixnumize (- min)) nargs)))
+             (inst cmp :dword temp (fixnumize (- max min)))
              (inst jmp :a err-lab))
             (t
              (cond ((= min 1)
-                    (inst test nargs nargs)
+                    (inst test :dword nargs nargs)
                     (inst jmp :e err-lab))
                    ((plusp min)
-                    (inst cmp nargs (fixnumize min))
+                    (inst cmp :dword nargs (fixnumize min))
                     (inst jmp :b err-lab))))))))
 
 ;; Signal an error about an untagged number.
