@@ -51,7 +51,7 @@
 ;; ~17 instructions vs. 35
 (define-vop ()
     (:policy :fast-safe)
-    (:translate wrapper-of)
+    (:translate layout-of)
     (:args (object :scs (descriptor-reg)))
     (:temporary (:sc unsigned-reg :offset rax-offset) rax)
     (:results (result :scs (descriptor-reg)))
@@ -81,16 +81,14 @@
       (inst movzx '(:byte :dword) rax object)
       LOAD-FROM-VECTOR
       (inst mov :dword result
-            (ea (make-fixup '**primitive-object-layouts**
-                           :symbol-value
+            (ea (make-fixup '**primitive-object-layouts** :symbol-value
                            (- (ash vector-data-offset word-shift)
                               other-pointer-lowtag))
                 nil rax 8)) ; no base register
       (inst jmp  done)
       NULL
       (inst mov  result (make-fixup 'null :layout))
-      DONE
-      #+metaspace (inst mov result (ea +5 result)))) ; layout->wrapper
+      DONE))
 (define-vop ()
     (:policy :fast-safe)
     (:translate %instanceoid-layout)
@@ -145,8 +143,7 @@
   (:generator 6
     (move temp data)
     (inst shl temp (- n-widetag-bits n-fixnum-tag-bits))
-    ;; merge in the widetag. We should really preserve bit 63 as well
-    ;; which could be a GC mark bit, but it's not concurrent at the moment.
+    ;; merge in the widetag
     (inst mov :byte temp (ea (- other-pointer-lowtag) x))
     (storew temp x 0 other-pointer-lowtag)))
 (define-vop (logior-header-bits)
@@ -519,3 +516,33 @@ number of CPU cycles elapsed as secondary value. EXPERIMENTAL."
    ;; Can't convert index to a code-relative index until the boxed header length
    ;; has been determined.
    (inst store-coverage-mark index)))
+
+(define-vop ()
+  (:translate sb-lockless::get-next)
+  (:policy :fast-safe)
+  (:args (node :scs (descriptor-reg)))
+  (:results (next-tagged :scs (descriptor-reg))
+            (next-bits :scs (descriptor-reg)))
+  (:generator 10
+    ;; Read the first user-data slot, which might be an untagged pointer
+    ;; to the next node. Conservative GC implicitly pins objects pointed to by untagged
+    ;; instance pointers now.
+    (loadw next-bits node (+ instance-slots-offset instance-data-start) instance-pointer-lowtag)
+    ;; Also returning the unadjusted bits as a secondary result.
+    ;; This can'be use the LEA instruction because _usually_ NEXT-BITS
+    ;; already hold a tagged pointer.
+    (inst mov next-tagged next-bits)
+    (inst or :byte next-tagged instance-pointer-lowtag)))
+
+(define-vop (switch-to-arena)
+  (:args (x :scs (descriptor-reg immediate)))
+  (:temporary (:sc unsigned-reg :offset rdi-offset :from (:argument 0)) rdi)
+  (:temporary (:sc unsigned-reg :offset rsi-offset) rsi)
+  (:vop-var vop)
+  (:ignore rsi)
+  (:generator 1
+    (inst mov rdi (if (sc-is x immediate) (tn-value x) x))
+    ;; We could have the vop declare that all C volatile regs are clobbered,
+    ;; but since this needs pseudo-atomic decoration anyway, it saves code size
+    ;; to make an assembly routine that preserves all registers.
+    (invoke-asm-routine 'call 'switch-to-arena vop)))
