@@ -7,15 +7,13 @@
 #include "runtime.h"
 #include "globals.h"
 #include "gc.h"
-#include "gc-internal.h"
-#include "gc-private.h"
-#include "gencgc-private.h"
 #include "lispregs.h"
 #include "genesis/arena.h"
 #include "genesis/gc-tables.h"
 #include "thread.h"
+#include "genesis/instance.h"
+#include "graphvisit.h"
 
-extern void acquire_gc_page_table_lock(), release_gc_page_table_lock();
 extern lispobj * component_ptr_from_pc(char *pc);
 
 // Arena memory block. At least one is associated with each arena,
@@ -24,7 +22,10 @@ struct arena_memblk {
     char* freeptr;
     char* limit;
     struct arena_memblk* next;
-    uword_t padding; // always 0
+    // The first memblk contains a 'struct arena' and a pthread mutex
+    // and the arena_memblk itself. Other memblks only have the memblk.
+    // Either way, Lisp objects commence at the 'allocator_base'.
+    char* allocator_base;
 };
 
 #if 1
@@ -57,7 +58,8 @@ lispobj sbcl_new_arena(size_t size)
     //   Memblk
     struct arena* arena = ARENA_GET_OS_MEMORY(size);
     memset(arena, 0, sizeof *arena);
-    arena->header = (sizeof (struct arena) / N_WORD_BYTES) << INSTANCE_LENGTH_SHIFT;
+    arena->header = ((sizeof (struct arena) / N_WORD_BYTES) << INSTANCE_LENGTH_SHIFT)
+      | INSTANCE_WIDETAG;
     struct arena_memblk* block =
       (void*)((char*)arena + ALIGN_UP(sizeof (struct arena), 2*N_WORD_BYTES));
     // arenas require threads, but the header for the mutex definition
@@ -78,20 +80,18 @@ lispobj sbcl_new_arena(size_t size)
     // all user allocations instead of having to skip the first batch
     // so that the arena struct always remains accessible.
     char* aligned_mem_base = PTR_ALIGN_UP(mem_base, CHUNK_ALIGN);
-    memset(mem_base, 0xCC, aligned_mem_base - mem_base);
-    block->freeptr = aligned_mem_base;
+    block->freeptr = block->allocator_base = aligned_mem_base;
     char *limit = (char*)arena + size;
     char *aligned_limit = PTR_ALIGN_DOWN(limit, CHUNK_ALIGN);
     block->limit = aligned_limit;
     block->next = NULL;
-    block->padding = 0;
     arena->uw_original_size = size;
     arena->uw_length = size;
     arena->uw_current_block = arena->uw_first_block = (uword_t)block;
     return make_lispobj(arena, INSTANCE_POINTER_LOWTAG);
 }
 
-static inline void* arena_mutex(struct arena* a) {
+static __attribute__((unused)) inline void* arena_mutex(struct arena* a) {
     return (void*)((char*)a + ALIGN_UP(sizeof (struct arena), 2*N_WORD_BYTES));
 }
 
@@ -103,6 +103,11 @@ static inline void* arena_mutex(struct arena* a) {
 void arena_release_memblks(lispobj arena_taggedptr)
 {
     struct arena* arena = (void*)native_pointer(arena_taggedptr);
+    // Avoid a race with GC in scavenge_arenas(). We're messing up the chain
+    // of blocks and can't have anything else looking at them.
+    __attribute__((unused)) lispobj old_hidden
+      = __sync_val_compare_and_swap(&arena->hidden, NIL, LISP_T);
+    gc_assert(old_hidden == NIL);
     ARENA_MUTEX_ACQUIRE(arena);
     struct arena_memblk* first = (void*)arena->uw_first_block;
     struct arena_memblk* block = first->next;
@@ -112,15 +117,22 @@ void arena_release_memblks(lispobj arena_taggedptr)
         block = next;
     }
     arena->uw_current_block = arena->uw_first_block;
-    char* mem_base = (char*)first + sizeof (struct arena_memblk);
-    first->freeptr = PTR_ALIGN_UP(mem_base, CHUNK_ALIGN);
+    first->freeptr = first->allocator_base;
     first->next = NULL;
-    arena->uw_extension_count = 0;
+    // Release huge-object blocks
+    block = (void*)arena->uw_huge_objects;
+    while (block) {
+        struct arena_memblk* next = block->next;
+        free(block);
+        block = next;
+    }
+    arena->uw_huge_objects = 0;
     arena->uw_length = arena->uw_original_size;
+    arena->hidden = NIL;
     ARENA_MUTEX_RELEASE(arena);
 }
 
-void AMD64_SYSV_ABI sbcl_delete_arena(lispobj arena_taggedptr)
+void sbcl_delete_arena(lispobj arena_taggedptr)
 {
     arena_release_memblks(arena_taggedptr);
     struct arena* arena = (void*)native_pointer(arena_taggedptr);
@@ -149,7 +161,16 @@ void AMD64_SYSV_ABI sbcl_delete_arena(lispobj arena_taggedptr)
     ARENA_DISPOSE_MEMORY(arena, arena->uw_original_size);
 }
 
-void AMD64_SYSV_ABI switch_to_arena(lispobj arena_taggedptr,
+static page_index_t close_heap_region(struct alloc_region* r, int page_type) {
+    page_index_t result = -1;
+    if (r->start_addr) {
+        result = find_page_index(r->start_addr);
+        gc_close_region(r, page_type);
+    }
+    return result;
+}
+
+void switch_to_arena(lispobj arena_taggedptr,
                      __attribute__((unused)) lispobj* ra) // return address
 {
     struct arena* arena = (void*)native_pointer(arena_taggedptr);
@@ -161,11 +182,12 @@ void AMD64_SYSV_ABI switch_to_arena(lispobj arena_taggedptr,
             (arena ? "to" : "from"),
             (arena ? arena : (void*)th->arena),
             (uword_t)ra, id);
-#endif
     struct extra_thread_data *extra_data = thread_extra_data(th);
+#endif
     if (arena) { // switching from the dynamic space to an arena
         if (th->arena)
-            lose("arena error: can't switch from %p to %p", (void*)th->arena, arena);
+            lose("arena error: can't switch from %p to %p", (void*)th->arena,
+                 (void*)arena_taggedptr);
         // Page table lock guards the arena chain, as well as the page table
         acquire_gc_page_table_lock();
         // See if this arena has ever been switched to,
@@ -175,9 +197,12 @@ void AMD64_SYSV_ABI switch_to_arena(lispobj arena_taggedptr,
             arena_chain = arena_taggedptr;
         }
         // Close only the non-system regions
-        if (th->mixed_tlab.start_addr) gc_close_region(&th->mixed_tlab, PAGE_TYPE_MIXED);
-        if (th->cons_tlab.start_addr) gc_close_region(&th->cons_tlab, PAGE_TYPE_CONS);
+        thread_extra_data(th)->mixed_page_hint =
+            close_heap_region(&th->mixed_tlab, PAGE_TYPE_MIXED);
+        thread_extra_data(th)->cons_page_hint =
+            close_heap_region(&th->cons_tlab, PAGE_TYPE_CONS);
         release_gc_page_table_lock();
+#if 0 // this causes a data race, the very thing it's trying to avoid
         int arena_index = fixnum_value(arena->index);
         /* If this thread has potentially used this arena previously, see if
          * the TLAB pointers can be restored based on token validity */
@@ -190,8 +215,10 @@ void AMD64_SYSV_ABI switch_to_arena(lispobj arena_taggedptr,
             }
             memset(state, 0, sizeof (arena_state));
         }
+#endif
     } else { // finished with the arena
         gc_assert(th->arena); // must have been an arena in use
+#if 0
         struct arena* old_arena = (void*)native_pointer(th->arena);
         int arena_index = fixnum_value(old_arena->index);
         // Ensure that this thread has enough space in its save area for the arena index.
@@ -212,6 +239,7 @@ void AMD64_SYSV_ABI switch_to_arena(lispobj arena_taggedptr,
         state->token = old_arena->uw_token;
         state->mixed = th->mixed_tlab;
         state->cons  = th->cons_tlab;
+#endif
         // Indicate that the tlabs have no space remaining.
         gc_set_region_empty(&th->mixed_tlab);
         gc_set_region_empty(&th->cons_tlab);
@@ -233,28 +261,62 @@ static void* memblk_claim_subrange(struct arena* a, struct arena_memblk* mem,
             if (mem != actual_current_block) { // oops - looking at the wrong memblk
                 mem = actual_current_block;
                 // fall into the mutex release and restart below
-            }  else {
-                // really add an extension block
-                if (a->uw_extension_count == a->uw_max_extensions) { // can't extend further
+            } else { // potentially extend
+                // An object occupying 1/512th or more of an extension is separately allocated.
+                int oversized = nbytes > (sword_t)(a->uw_growth_amount >> 9);
+                long request = oversized ? nbytes : (sword_t)a->uw_growth_amount;
+                if (a->uw_length + request > a->uw_size_limit) { // limit reached
                     ARENA_MUTEX_RELEASE(a);
-                    lose("Fatal: arena memory exhausted");
+                    lose("Fatal: won't add arena %s block. Length=%lx request=%lx max=%lx",
+                         oversized ? "huge-object" : "extension",
+                         (long)a->uw_length, (long)request, (long)a->uw_size_limit);
+                }
+                // For a huge object, ensure that there will be adequate space after aligning
+                // the bounds. Don't worry about it for a regular extension block though
+                // because there's no chance that the current request won't fit.
+                // Moreover, use malloc() for oversized objects instead of the macro
+                // because we don't actually remember the originally requested size
+                // to pass to ARENA_DISPOSE_MEMORY in arena_release_memblks
+                long actual_request;
+                char* new_mem;
+                if (oversized) {
+                    actual_request = request + 3*CHUNK_ALIGN;
+                    new_mem = malloc(actual_request);
+                } else {
+                    actual_request = request;
+                    new_mem = ARENA_GET_OS_MEMORY(actual_request);
                 }
-                char* new_mem = ARENA_GET_OS_MEMORY(a->uw_growth_amount);
                 if (new_mem == 0) {
                     ARENA_MUTEX_RELEASE(a);
                     lose("Fatal: arena memory exhausted and could not obtain more memory");
                 }
                 struct arena_memblk* extension= (void*)new_mem;
                 char* mem_base = new_mem + sizeof (struct arena_memblk);
+                // Alignment serves the needs of hide/unhide because mprotect()
+                // operates only on boundaries as dictated by the OS and/or CPU.
                 char* aligned_mem_base = PTR_ALIGN_UP(mem_base, CHUNK_ALIGN);
-                extension->freeptr = aligned_mem_base;
-                char* limit = new_mem + a->uw_growth_amount;
+                extension->freeptr = extension->allocator_base = aligned_mem_base;
+                char* limit = new_mem + actual_request;
                 char* aligned_limit = PTR_ALIGN_DOWN(limit, CHUNK_ALIGN);
                 extension->limit = aligned_limit;
                 extension->next = NULL;
-                extension->padding = 0;
-                a->uw_length += a->uw_growth_amount; // tally up the total length
-                a->uw_extension_count++;
+                // tally up the total length
+                // For huge objects, count only the directly requested space,
+                // not the "bookends" (alignment chunks)
+                a->uw_length += request;
+                if (oversized) {
+                    long usable_space = aligned_limit - aligned_mem_base;
+                    // This should assert() because it's very bad, but assertions can be
+                    // disabled, so explicitly lose() if it happens.
+                    if (nbytes > usable_space) lose("alignment glitch");
+                    extension->next = (void*)a->uw_huge_objects;
+                    a->uw_huge_objects = (uword_t)extension;
+                    // A huge object needs to be zero-filled. It probably was, because it
+                    // probably just got mmapped() but there's no way to know.
+                    memset(aligned_mem_base, 0, nbytes);
+                    ARENA_MUTEX_RELEASE(a);
+                    return aligned_mem_base;
+                }
                 mem->next = extension;
                 // Other threads can start using the new block already
                 a->uw_current_block = (lispobj)extension;
@@ -294,11 +356,16 @@ lispobj* handle_arena_alloc(struct thread* th, struct alloc_region* region,
      *  1. amount remaining in the TLAB is not significant. Just toss it out.
      *     Refill the TLAB with the default quantum plus the user's actual request.
      *     Then carve out the new object from the beginning of the TLAB.
+     *     Exception: if 'nbytes' exceeds 64KiB then always treat it as case 2.
+     *     This avoids an anomaly in the situation where 'nbytes' so large that it
+     *     can not be obtained from the current arena block, but instead uses malloc()
+     *     directly. In such case, attempting to refill the TLAB from the same
+     *     block is futile.
      *
      *  2. amount remaining in the TLAB is worth keeping.
      *     Don't refill the TLAB; just make the new object as a discrete claim.
      */
-    if (avail < min_keep) { // case 1
+    if (avail < min_keep && nbytes <= 65536) { // case 1
         struct arena* in_use_arena = (void*)native_pointer(th->arena);
         __sync_fetch_and_add(&in_use_arena->uw_bytes_wasted, avail);
         long total_request = nbytes + 8192;
@@ -320,8 +387,11 @@ long arena_bytes_used(lispobj arena_taggedptr)
     ARENA_MUTEX_ACQUIRE(arena);
     struct arena_memblk* block = (void*)arena->uw_first_block;
     do {
-        sum += block->freeptr - (char*)block;
+        sum += block->freeptr - block->allocator_base;
     } while ((block = block->next) != NULL);
+    for ( block = (void*)arena->uw_huge_objects ; block ; block = block->next) {
+        sum += block->limit - block->allocator_base;
+    }
     ARENA_MUTEX_RELEASE(arena);
     return sum;
 }
@@ -347,12 +417,27 @@ void gc_scavenge_arenas()
             if (a->hidden == NIL) {
                 struct arena_memblk* block = (void*)a->uw_first_block;
                 do {
-                    // The block is its own lower bound for scavenge.
-                    // Its first 4 words look like fixnums, so no need to skip 'em.
-                    fprintf(stderr, "Arena @ %p: scavenging %p..%p\n",
-                            a, block, block->freeptr);
-                    heap_scavenge((lispobj*)block, (lispobj*)block->freeptr);
+                    if (gencgc_verbose)
+                        fprintf(stderr, "Arena @ %p: scavenging %p..%p\n",
+                                a, block->allocator_base, block->freeptr);
+#ifdef LISP_FEATURE_MARK_REGION_GC
+                    mr_trace_bump_range((lispobj*)block->allocator_base, (lispobj*)block->freeptr);
+#else
+                    heap_scavenge((lispobj*)block->allocator_base, (lispobj*)block->freeptr);
+#endif
                 } while ((block = block->next) != NULL);
+                for ( block = (void*)a->uw_huge_objects ; block ; block = block->next ) {
+                    lispobj* obj = (lispobj*)block->allocator_base;
+#ifdef LISP_FEATURE_MARK_REGION_GC
+                    // fprintf(stderr, "arena huge object @ %p: %lx %lx\n", obj, obj[0], obj[1]);
+                    // probably could call trace_other_object directly here?
+                    lispobj* end = obj + object_size(obj);
+                    mr_trace_bump_range(obj, end);
+#else
+                    lispobj header = *obj;
+                    scavtab[header_widetag(header)](obj, header);
+#endif
+                }
             }
             chain = a->link;
         } while (chain != NIL);
@@ -379,7 +464,7 @@ lispobj find_containing_arena(lispobj ptr) {
 }
 
 static lispobj target_arena;
-static inline boolean interesting_arena_pointer_p(lispobj ptr)
+static inline int interesting_arena_pointer_p(lispobj ptr)
 {
     lispobj arena = is_lisp_pointer(ptr) ? find_containing_arena(ptr) : 0;
     if (!arena) return 0; // uninteresting
@@ -463,14 +548,18 @@ int find_dynspace_to_arena_ptrs(lispobj arena, lispobj result_buffer)
         /* This produces false positives, don't return potential pointers, but instead print them.
          * Using the output, you can try to probe the suspect memory with SB-VM:HEXDUMP */
         if (th == get_sb_vm_thread()) {
-          scan_thread_control_stack(&arena, // = the approximate stack pointer
-                                    th->control_stack_end,
-                                    th->lisp_thread);
+            scan_thread_control_stack(&arena, // = the approximate stack pointer
+                                      th->control_stack_end,
+                                      th->lisp_thread);
         } else {
-          int ici = fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX, th));
-          if (ici != 1) lose("can't find interrupt context");
-          lispobj sp = *os_context_register_addr(nth_interrupt_context(0, th), reg_SP);
-          scan_thread_control_stack((lispobj*)sp, th->control_stack_end, th->lisp_thread);
+#ifdef LISP_FEATURE_SB_SAFEPOINT
+            lispobj *sp = os_get_csp(th);
+#else
+            int ici = fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX, th));
+            if (ici != 1) lose("can't find interrupt context");
+            lispobj sp = *os_context_register_addr(nth_interrupt_context(0, th), reg_SP);
+#endif
+            scan_thread_control_stack((lispobj*)sp, th->control_stack_end, th->lisp_thread);
         }
 #endif
         scan_thread_words((lispobj*)th->binding_stack_start,
@@ -491,9 +580,40 @@ int find_dynspace_to_arena_ptrs(lispobj arena, lispobj result_buffer)
     return result;
 }
 
+static int count_arena_objects(lispobj arena, int *pnchunks)
+{
+    struct arena* a = (void*)native_pointer(arena);
+    struct arena_memblk* blk = (void*)a->uw_first_block;
+    int nobjects = 0;
+    int nchunks = 0;
+    do {
+        ++nchunks;
+        lispobj* where = (void*)ALIGN_UP(((uword_t)blk + sizeof (struct arena_memblk)),
+                                         CHUNK_ALIGN);
+        lispobj* limit = (void*)blk->freeptr;
+        while (where < limit) {
+            if (*where == (uword_t)-1) { // filler
+                where += 2;
+            } else {
+                ++nobjects;
+                where += object_size(where);
+            }
+        }
+        blk = blk->next;
+    } while (blk);
+    *pnchunks = nchunks;
+    return nobjects;
+}
+
 void arena_mprotect(lispobj arena, int option)
 {
 #ifndef LISP_FEATURE_WIN32
+    if (option) { // protecting
+        int nchunks;
+        int count = count_arena_objects(arena, &nchunks); // mainly as a sanity-check
+        fprintf(stderr, "arena_mprotect %p: %d objects in %d chunk(s)\n",
+                (void*)arena, count, nchunks);
+    }
     // PROT_EXEC is not needed. Code blobs always go to dynamic or immobile space
     int prot = option ? PROT_NONE : (PROT_READ|PROT_WRITE /*|PROT_EXEC*/);
     struct arena* a = (void*)native_pointer(arena);
@@ -505,6 +625,7 @@ void arena_mprotect(lispobj arena, int option)
         // the block itself is not within [base,limit] and so can be read even if prot==PROT_NONE
         blk = blk->next;
     } while (blk);
+    // TODO: mprotect huge-object blocks
 #endif
 }
 
@@ -515,8 +636,8 @@ lispobj arena_find_containing_object(lispobj arena, char* ptr)
     do {
         lispobj* where = (void*)ALIGN_UP(((uword_t)blk + sizeof (struct arena_memblk)),
                                          CHUNK_ALIGN);
-        lispobj* limit = (void*)blk->limit;
-        while (where <  limit) {
+        lispobj* limit = (void*)blk->freeptr;
+        while (where < limit) {
             if (*where == (uword_t)-1) { // filler
                 where += 2;
             } else {
@@ -534,20 +655,61 @@ lispobj arena_find_containing_object(lispobj arena, char* ptr)
 int diagnose_arena_fault(os_context_t* context, char *addr)
 {
 #ifndef LISP_FEATURE_WIN32
+    if (!arena_chain) return 0; // not handled
     lispobj arena = find_containing_arena((lispobj)addr);
-    if (!arena) return 0; // not handled
-    if (arena && ((struct arena*)native_pointer(arena))->hidden == LISP_T) {
-        arena_mprotect(arena, 0); // unprotect it and find the object
-        lispobj obj = arena_find_containing_object(arena, addr);
-        if (obj) {
-            fprintf(stderr, "access @ %p sees hidden arena object @ %p in arena %p\n",
-                    addr, (void*)obj, (void*)arena);
-            fflush(stderr);
-        }
-        //        arena_mprotect(arena, 1); // put it back the way it was
-        lisp_memory_fault_error(context, addr);
-        return 1;
+    struct thread* th = get_sb_vm_thread();
+    struct thread_instance* instance = (void*)native_pointer(th->lisp_thread);
+    lispobj name = instance->_name;
+    char *c_string = 0;
+    if (name != NIL) {
+        struct vector* string = (void*)native_pointer(name);
+        if (header_widetag(string->header) == SIMPLE_BASE_STRING_WIDETAG)
+            c_string = (char*)string->data;
     }
-    return 0;
+    fprintf(stderr, "trying diagnose_arena_fault(%p,'%s')\n",
+            addr, c_string);
+    if (!arena) {
+        fprintf(stderr, "fault is not in an arena\n");
+        fflush(stderr);
+        return 0; // not handled
+    }
+    int hidden = ((struct arena*)native_pointer(arena))->hidden == LISP_T;
+    fprintf(stderr, "fault in arena %p [%s]\n", (void*)arena, (hidden ? "HIDDEN" : "visible"));
+    fflush(stderr);
+    if (!hidden) return 0;
+    arena_mprotect(arena, 0); // unprotect it and find the object
+    fprintf(stderr, "unprotected OK\n");
+    fflush(stderr);
+    lispobj obj = arena_find_containing_object(arena, addr);
+    if (!obj) {
+        fprintf(stderr, "could not find containing lispobj\n");
+        fflush(stderr);
+        return 0;
+    }
+    fprintf(stderr, "access of object @ %p\n", (void*)obj);
+    fflush(stderr);
+    lisp_memory_fault_error(context, addr);
 #endif
+    return 1;
+}
+
+struct visitor {
+    lispobj arena;
+    long nwords;
+};
+
+static void visit(lispobj obj, void* arg) {
+    struct visitor* v = arg;
+    if (find_containing_arena(obj) == v->arena) v->nwords += object_size(native_pointer(obj));
+}
+size_t count_arena_live_bytes(lispobj arena) {
+    struct hopscotch_table h;
+    struct visitor v;
+    v.arena = arena;
+    v.nwords = 0;
+    struct grvisit_context* c =
+        visit_heap_from_static_roots(&h, visit, &v);
+    hopscotch_destroy(&h);
+    free(c);
+    return v.nwords * N_WORD_BYTES;
 }
