@@ -18,7 +18,6 @@
 #include <sys/param.h>
 #include <sys/file.h>
 #include "sbcl.h"
-#include "./signal.h"
 #include "os.h"
 #include "arch.h"
 #include "globals.h"
@@ -71,6 +70,12 @@ os_context_sigmask_addr(os_context_t *context)
     return &(context->uc_sigmask);
 }
 
+os_context_register_t *
+os_context_flags_addr(os_context_t *context)
+{
+    return (os_context_register_t *)&(context->uc_mcontext.pstate);
+}
+
 void
 os_restore_fp_control(os_context_t *context)
 {
