@@ -37,7 +37,7 @@
 #include "genesis/static-symbols.h"
 #include "genesis/primitive-objects.h"
 #include "genesis/gc-tables.h"
-#include "gc-internal.h"
+#include "gc.h"
 #include "../../tlsf-bsd/tlsf/tlsf.h"
 extern void* tlsf_control;
 
@@ -497,13 +497,22 @@ dump_cmd(char **ptr)
                    visible(cptr[0]), visible(cptr[1]),
                    visible(cptr[2]), visible(cptr[3]));
 #endif
-#ifdef LISP_FEATURE_GENCGC
+#ifdef LISP_FEATURE_GENERATIONAL
             if (aligned) {
                 lispobj ptr = *(lispobj*)addr;
                 int gen;
                 if (is_lisp_pointer(ptr) && gc_managed_heap_space_p(ptr)
-                    && (gen = gc_gen_of(ptr, 99)) != 99) // say that static is 99
+                    && (gen = gc_gen_of(ptr, 99)) != 99) { // say that static is 99
                     if (gen != 99) printf(" | %d", gen);
+                } else {
+                    printf("    "); // padding to make MR part line up
+                }
+            }
+#endif
+#ifdef LISP_FEATURE_MARK_REGION_GC
+            if (aligned && find_page_index(addr) != -1) {
+                extern bool allocation_bit_marked(void*);
+                printf(" %c", allocation_bit_marked(addr) ? '*' : ' ');
             }
 #endif
             if (decode && addr == (char*)next_object) {
@@ -566,17 +575,13 @@ regs_cmd(char __attribute__((unused)) **ptr)
     printf("\n");
 #endif
 
-#ifdef LISP_FEATURE_GENCGC
+#ifdef LISP_FEATURE_GENERATIONAL
     /* printf("DYNAMIC\t=\t%p\n", (void*)DYNAMIC_SPACE_START); */
 #else
     printf("STATIC\t=\t%p   ", static_space_free_pointer);
     printf("RDONLY\t=\t%p   ", read_only_space_free_pointer);
     printf("DYNAMIC\t=\t%p\n", (void*)current_dynamic_space);
 #endif
-
-#ifndef LISP_FEATURE_GENCGC
-    printf("TRIGGER\t=\t%p\n", (void*)current_auto_gc_trigger);
-#endif
     return 0;
 }
 
@@ -864,8 +869,8 @@ void gc_stop_the_world() { } // do nothing
 void gc_start_the_world() { } // do nothing
 #include <errno.h>
 #include "core.h"
-#include "gencgc-private.h"
 struct lisp_startup_options lisp_startup_options;
+char gc_coalesce_string_literals;
 
 void unwind_binding_stack() { lose("Can't unwind binding stack"); }
 FILE *prepare_to_save(__attribute__((unused)) char *filename,
