@@ -18,19 +18,12 @@
 #include "os.h"
 #include "interrupt.h"
 #include "lispregs.h"
-#include "signal.h"
+#include <signal.h>
 #include "interrupt.h"
 #include "interr.h"
 #include "breakpoint.h"
-#include "alloc.h"
-
-#if defined(LISP_FEATURE_GENCGC)
-#include "gencgc-alloc-region.h"
-#endif
-
-#ifdef LISP_FEATURE_SB_THREAD
-#include "getallocptr.h"
-#endif
+#include "pseudo-atomic.h"
+#include "gc-assert.h"
 
   /* The header files may not define PT_DAR/PT_DSISR.  This definition
      is correct for all versions of ppc linux >= 2.0.30
@@ -45,10 +38,8 @@
 
      Caveat callers.  */
 
-#if defined (LISP_FEATURE_DARWIN) || defined(LISP_FEATURE_LINUX)
 #ifndef PT_DAR
 #define PT_DAR          41
-#endif
 
 #ifndef PT_DSISR
 #define PT_DSISR        42
@@ -79,69 +70,26 @@ arch_get_bad_addr(int sig, siginfo_t *code, os_context_t *context)
 void
 arch_skip_instruction(os_context_t *context)
 {
-    char** pcptr;
-    pcptr = (char**) os_context_pc_addr(context);
-    *pcptr += 4;
+    OS_CONTEXT_PC(context) += 4;
 }
 
 unsigned char *
 arch_internal_error_arguments(os_context_t *context)
 {
-    return (unsigned char *)(*os_context_pc_addr(context)+4);
+    return (unsigned char *)(OS_CONTEXT_PC(context)+4);
 }
 
 
-boolean
-arch_pseudo_atomic_atomic(os_context_t *context)
-{
-#ifdef LISP_FEATURE_SB_THREAD
-    struct thread *thread = arch_os_get_current_thread();
-
-    if (foreign_function_call_active_p(thread)) {
-        return get_pseudo_atomic_atomic(thread);
-    } else return
-#else
-    /* FIXME: this foreign_function_call_active test is dubious at
-     * best. If a foreign call is made in a pseudo atomic section
-     * (?) or more likely a pseudo atomic section is in a foreign
-     * call then an interrupt is executed immediately. Maybe it
-     * has to do with C code not maintaining pseudo atomic
-     * properly. MG - 2005-08-10
-     *
-     * The foreign_function_call_active used to live at each call-site
-     * to arch_pseudo_atomic_atomic, but this seems clearer.
-     * --NS 2007-05-15 */
-    return (!foreign_function_call_active_p(arch_os_get_current_thread())) &&
-#endif
-        ((*os_context_register_addr(context,reg_ALLOC)) & flag_PseudoAtomic);
+bool arch_pseudo_atomic_atomic(struct thread *thread) {
+    return get_pseudo_atomic_atomic(thread);
 }
 
-void
-arch_set_pseudo_atomic_interrupted(os_context_t *context)
-{
-#ifdef LISP_FEATURE_SB_THREAD
-    struct thread *thread = arch_os_get_current_thread();
-
-    if (foreign_function_call_active_p(thread)) {
-        set_pseudo_atomic_interrupted(thread);
-    } else
-#endif
-        *os_context_register_addr(context,reg_ALLOC)
-            |= flag_PseudoAtomicInterrupted;
+void arch_set_pseudo_atomic_interrupted(struct thread *thread) {
+    set_pseudo_atomic_interrupted(thread);
 }
 
-void
-arch_clear_pseudo_atomic_interrupted(os_context_t *context)
-{
-#ifdef LISP_FEATURE_SB_THREAD
-    struct thread *thread = arch_os_get_current_thread();
-
-    if (foreign_function_call_active_p(thread)) {
-        clear_pseudo_atomic_interrupted(thread);
-    } else
-#endif
-        *os_context_register_addr(context,reg_ALLOC)
-            &= ~flag_PseudoAtomicInterrupted;
+void arch_clear_pseudo_atomic_interrupted(struct thread *thread) {
+    clear_pseudo_atomic_interrupted(thread);
 }
 
 unsigned int
@@ -181,7 +129,7 @@ arch_remove_breakpoint(void *pc, unsigned int orig_inst)
 static unsigned int *skipped_break_addr, displaced_after_inst;
 static sigset_t orig_sigmask;
 
-static boolean
+static bool
 should_branch(os_context_t *context, unsigned int orig_inst)
 {
     /* orig_inst is a conditional branch instruction.  We need to
@@ -208,7 +156,7 @@ arch_do_displaced_inst(os_context_t *context, unsigned int orig_inst)
 {
     /* not sure how we ensure that we get the breakpoint reinstalled
      * after doing this -dan */
-    unsigned int *pc = (unsigned int *)(*os_context_pc_addr(context));
+    unsigned int *pc = (unsigned int *)OS_CONTEXT_PC(context);
     unsigned int *next_pc;
     int op = orig_inst >> 26;
     int sub_op = (orig_inst & 0x7fe) >> 1;  /* XL-form sub-opcode */
@@ -262,9 +210,9 @@ arch_do_displaced_inst(os_context_t *context, unsigned int orig_inst)
 }
 
 #define INLINE_ALLOC_DEBUG 0
-#ifdef LISP_FEATURE_CHENEYGC
-#define handle_allocation_trap(x) (0)
-#else
+#define ALLOC_TRAP_LISTIFY 1
+#define ALLOC_TRAP_CONS    2
+#define ALLOC_TRAP_GENERAL 3
 /*
  * Return non-zero if the current instruction is an allocation trap
  */
@@ -279,25 +227,29 @@ allocation_trap_p(os_context_t * context)
      * |31| TO|dst|src|  4|0|  TW - trap word
      * |31| TO|dst|src| 68|0|  TD - trap doubleword
      *
-     *   TO = #b00001 for LGT
-     *        #b00101 for LGE
+     *   TO = #b00100 for EQ
+     *        #b00001 for LGT (logical greater-than)
+     *        #b00101 for LGE (logical greater-or-equal)
+     *        #b00010 for LLT (logical less-than)
+     *        #b00110 for LLE (logical less-or-equal)
      */
-    unsigned *pc = (unsigned int *) *os_context_pc_addr(context);
+    unsigned *pc = (unsigned int *)OS_CONTEXT_PC(context);
     unsigned inst = *pc;
     unsigned opcode = inst >> 26;
     unsigned src = (inst >> 11) & 0x1f;
-    // unsigned dst = (inst >> 16) & 0x1f;
+    unsigned dst = (inst >> 16) & 0x1f;
     unsigned to = (inst >> 21) & 0x1f;
     unsigned subcode = inst & 0x7ff;
 
-    if (opcode == 31 && (to == 1 || to == 5) && src == reg_NL3
+    // recognize the listify-rest-args allocation trap
+    if (opcode == 31 && to == 5 && src == dst && subcode == 4<<1)
+        return ALLOC_TRAP_LISTIFY;
+
+    // FIXME: we can remove the wired use of NL3 in the allocator
+    if (opcode == 31 && (to == 1 || to == 2)
+        && (src == reg_NL3 || dst == reg_NL3)
         && (subcode == 4<<1 || subcode == 68<<1)) {
-        /* It doesn't much matter which trap option we pick for "could be large"
-         * but I've chosen "TGE" because of the choices, that one is subject to spurious
-         * failure, and I'd prefer not to spuriously fail on a 2-word cons.
-         * (Spurious failure occurs when the EQ condition is met, meaning the allocation
-         * would have worked, but the trap happens regardless) */
-        int success = (to == 5) ? 1 : -1; // 1 = single-object page ok, -1 = not ok
+        int success = (to == 2) ? ALLOC_TRAP_CONS : ALLOC_TRAP_GENERAL;
 
         /*
          * We got the instruction.  Now, look back to make sure it was
@@ -310,7 +262,7 @@ allocation_trap_p(os_context_t * context)
         opcode = add_inst >> 26;
         if ((opcode == 31) && (266 == ((add_inst >> 1) & 0x1ff))) {
             return success;
-        } else if ((opcode == 14)) {
+        } else if (opcode == 14) {
             return success;
         } else {
             fprintf(stderr,
@@ -321,124 +273,73 @@ allocation_trap_p(os_context_t * context)
     return 0;
 }
 
-#ifndef boxed_region
-#define boxed_region gc_alloc_region[0]
-#endif
-
 static int
 handle_allocation_trap(os_context_t * context)
 {
-    int alloc_trap_p = allocation_trap_p(context);
+    int alloc_trap_kind = allocation_trap_p(context);
+
+    if (!alloc_trap_kind) return 0;
+
+    struct thread* thread = get_sb_vm_thread();
+    gc_assert(!foreign_function_call_active_p(thread));
+    if (gencgc_alloc_profiler && thread->state_word.sprof_enable)
+        record_backtrace_from_context(context, thread);
 
-    if (!alloc_trap_p) return 0;
-    gc_assert(!foreign_function_call_active_p(arch_os_get_current_thread()));
     fake_foreign_function_call(context);
+    struct interrupt_data *data = &thread_interrupt_data(thread);
+    data->allocation_trap_context = context;
 
-    unsigned int *pc = (unsigned int *) (*os_context_pc_addr(context));
+    unsigned int *pc = (unsigned int *)OS_CONTEXT_PC(context);
 
-    /*
-     * Go back and look at the add/addi instruction.  The second src arg
-     * is the size of the allocation.  Get it and call alloc to allocate
-     * new space.
-     */
-
-    unsigned int inst = pc[-1];
-    int target = (inst >> 21) & 0x1f;
-    unsigned int opcode = inst >> 26;
-#if INLINE_ALLOC_DEBUG
-    fprintf(stderr, "  add inst  = 0x%08x, opcode = %d\n", inst, opcode);
-#endif
-    sword_t size = 0;
-    if (opcode == 14) {
+    if (alloc_trap_kind == ALLOC_TRAP_LISTIFY) {
+        unsigned inst = pc[0];
+        int count_reg = (inst >> 11) & 0x1f;
+        // There's a dummy trap instruction which encodes the context and result.
+        inst = pc[1];
+        unsigned context_reg = (inst >> 16) & 0x1f;
+        unsigned result_reg = (inst >> 11) & 0x1f;
+        lispobj* argv = (void*)*os_context_register_addr(context, context_reg);
+        lispobj nbytes = *os_context_register_addr(context, count_reg);
+        extern lispobj listify_rest_arg(lispobj*, sword_t);
+        lispobj result = listify_rest_arg(argv, nbytes);
+        *os_context_register_addr(context, result_reg) = result;
+        // Skip this and the next instruction
+        OS_CONTEXT_PC(context) += 8;
+    } else {
         /*
-         * ADDI temp-tn, alloc-tn, size
-         *
-         * Extract the size
+         * Go back and look at the add/addi instruction.  The second src arg
+         * is the size of the allocation.  Get it and call alloc to allocate
+         * new space.
+         * (Alternatively we could look at the trap instruction to see which
+         * register was compared against the region free pointer. Subtracting
+         * the region base would yield the size)
          */
-        size = (inst & 0xffff);
-    } else if (opcode == 31) {
-        /*
-         * ADD temp-tn, alloc-tn, size-tn
-         *
-         * Extract the size
-         */
-        int reg;
-
-        reg = (inst >> 11) & 0x1f;
-#if INLINE_ALLOC_DEBUG
-        fprintf(stderr, "  add, reg = %s\n", lisp_register_names[reg]);
-#endif
-        size = *os_context_register_addr(context, reg);
-
-    }
-
-#if INLINE_ALLOC_DEBUG
-    fprintf(stderr, "Alloc %d to %s\n", size, lisp_register_names[target]);
-    if ((((unsigned long)boxed_region.end_addr + size) / GENCGC_CARD_BYTES) ==
-        (((unsigned long)boxed_region.end_addr) / GENCGC_CARD_BYTES)) {
-      fprintf(stderr,"*** possibly bogus trap allocation of %d bytes at %p\n",
-              size, (void*)target_ptr);
-      fprintf(stderr, "    dynamic_space_free_pointer: %p, boxed_region.end_addr %p\n",
-              dynamic_space_free_pointer, boxed_region.end_addr);
-    }
-    fprintf(stderr, "Ready to alloc\n");
-    fprintf(stderr, "free_pointer = %p\n", dynamic_space_free_pointer);
-#endif
-
-    /*
-     * alloc-tn was incremented by size.  Need to decrement it by size
-     * to restore its original value. This is not true on GENCGC
-     * anymore. d_s_f_p and reg_alloc get out of sync, but the p_a
-     * bits stay intact and we set it to the proper value when it
-     * needs to be. Keep this comment here for the moment in case
-     * somebody tries to figure out what happened here.
-     */
-    /*    dynamic_space_free_pointer =
-        (lispobj *) ((long) dynamic_space_free_pointer - size);
-    */
-
-    char *memory;
-    {
+        unsigned int inst = pc[-1];
+        int target = (inst >> 21) & 0x1f;
+        unsigned int opcode = inst >> 26;
+        sword_t size = 0;
+        if (opcode == 14) {        // ADDI temp-tn, alloc-tn, size
+            size = (inst & 0xffff);
+        } else if (opcode == 31) { // ADD temp-tn, alloc-tn, size-tn
+            int reg;
+            reg = (inst >> 11) & 0x1f;
+            size = *os_context_register_addr(context, reg);
+        }
+        char *memory;
         extern lispobj *alloc(sword_t), *alloc_list(sword_t);
-        struct interrupt_data *data =
-            arch_os_get_current_thread()->interrupt_data;
-        data->allocation_trap_context = context;
-        memory = (char*)(alloc_trap_p < 0 ? alloc_list(size) : alloc(size));
-        data->allocation_trap_context = 0;
+        memory = (char*)(alloc_trap_kind==ALLOC_TRAP_CONS ? alloc_list(size) : alloc(size));
+        // ALLOCATION wants the result to point to the end of the object!
+        *os_context_register_addr(context, target) =
+          (os_context_register_t)(memory + size);
+        // Skip 2 instructions: the trap, and the writeback of free pointer
+        OS_CONTEXT_PC(context) = (uword_t)(pc + 2); // ('pc' is of type int*)
     }
 
-#if INLINE_ALLOC_DEBUG
-    fprintf(stderr, "alloc returned %p\n", memory);
-    fprintf(stderr, "free_pointer = %p\n", dynamic_space_free_pointer);
-#endif
-
-    /*
-     * The allocation macro wants the result to point to the end of the
-     * object!
-     */
-    memory += size;
-
-#if INLINE_ALLOC_DEBUG
-    fprintf(stderr, "object end at %p\n", memory);
-#endif
-
-    *os_context_register_addr(context, target) = (unsigned long) memory;
-#ifndef LISP_FEATURE_SB_THREAD
-    /* This is handled by the fake_foreign_function_call machinery on
-     * threaded targets. */
-    *os_context_register_addr(context, reg_ALLOC) =
-      (unsigned long) dynamic_space_free_pointer
-      | (*os_context_register_addr(context, reg_ALLOC)
-         & LOWTAG_MASK);
-#endif
-
+    data->allocation_trap_context = 0;
     undo_fake_foreign_function_call(context);
 
-    // Skip 2 instructions: the trap, and the writeback of free pointer
-    (*os_context_pc_addr(context)) = (uword_t)(pc + 2);
     return 1; // handled
 }
-#endif
 
 #if defined LISP_FEATURE_SB_THREAD
 static int
@@ -464,10 +365,10 @@ handle_tls_trap(os_context_t * context, uword_t pc, unsigned int code)
      *
      */
 
-    boolean handle_it = 0;
+    bool handle_it = 0;
     unsigned prev_inst;
     if ((code & ~(31 << 16)) == ((3<<26)|(4<<21))) { // mask out RA for test
-        prev_inst= ((u32*)pc)[-1];
+        prev_inst= ((uint32_t*)pc)[-1];
         int16_t disp = (prev_inst & 0xFFFF);
         handle_it = (prev_inst >> 26) == 32 // a load
           // RT of the load must be RA of the trap
@@ -476,7 +377,7 @@ handle_tls_trap(os_context_t * context, uword_t pc, unsigned int code)
     }
     if (!handle_it) return 0;
 
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     set_pseudo_atomic_atomic(thread);
 
     int symbol_reg = (prev_inst >> 16) & 31;
@@ -534,7 +435,7 @@ arch_handle_breakpoint(os_context_t *context)
 void
 arch_handle_fun_end_breakpoint(os_context_t *context)
 {
-    *os_context_pc_addr(context) = (uword_t)handle_fun_end_breakpoint(context);
+    OS_CONTEXT_PC(context) = (uword_t)handle_fun_end_breakpoint(context);
 }
 
 void
@@ -544,17 +445,17 @@ arch_handle_after_breakpoint(os_context_t *context)
     os_flush_icache((os_vm_address_t) skipped_break_addr,
                     sizeof(unsigned int));
     skipped_break_addr = NULL;
-    *(unsigned int *)*os_context_pc_addr(context)
-        = displaced_after_inst;
+    // This writes an instruction, NOT assigns to the pc.
+    *(unsigned int *)OS_CONTEXT_PC(context) = displaced_after_inst;
     *os_context_sigmask_addr(context)= orig_sigmask;
-    os_flush_icache((os_vm_address_t) *os_context_pc_addr(context),
+    os_flush_icache((os_vm_address_t) OS_CONTEXT_PC(context),
                     sizeof(unsigned int));
 }
 
 void
 arch_handle_single_step_trap(os_context_t *context, int trap)
 {
-    unsigned int code = *((u32 *)(*os_context_pc_addr(context)));
+    unsigned int code = *(uint32_t *)OS_CONTEXT_PC(context);
     int register_offset = code >> 8 & 0x1f;
     handle_single_step_trap(context, trap, register_offset);
     arch_skip_instruction(context);
@@ -563,7 +464,6 @@ arch_handle_single_step_trap(os_context_t *context, int trap)
 #if 0
 static void dump_cpu_state(char *reason, os_context_t* context)
 {
-    extern void sigset_tostring(const sigset_t *sigset, char* result, int result_length);
     int i,j,r=0;
     sigset_t cur_sigset;
     char buf[100];
@@ -573,7 +473,7 @@ static void dump_cpu_state(char *reason, os_context_t* context)
     pthread_sigmask(0, 0, &cur_sigset); sigset_tostring(&cur_sigset, buf, sizeof buf);
     fprintf(stderr, " curmask=%s\n", buf);
     fprintf(stderr, "  $pc=%16lx  $lr=%16lx $ctr=%16lx  $cr=%16lx\n",
-            *os_context_pc_addr(context),
+            OS_CONTEXT_PC(context),
             *os_context_lr_addr(context),
             *os_context_ctr_addr(context),
             *os_context_cr_addr(context));
@@ -590,19 +490,19 @@ static void dump_cpu_state(char *reason, os_context_t* context)
 static void
 sigtrap_handler(int signal, siginfo_t *siginfo, os_context_t *context)
 {
-  uword_t pc = *os_context_pc_addr(context);
-  unsigned int code = *(u32*)pc;
+  uword_t pc = OS_CONTEXT_PC(context);
+  unsigned int code = *(uint32_t*)pc;
 
 #ifdef LISP_FEATURE_SIGILL_TRAPS
     if (signal == SIGILL) {
         if (code == 0x7C0002A6) { // allocation region overflow trap
             // there is an actual trap instruction located 2 instructions later.
             // pretend the trap happened there.
-            *os_context_pc_addr(context) = pc + 8;
+            OS_CONTEXT_PC(context) = pc + 8;
             if (handle_allocation_trap(context)) return;
         }
         if (code == 0x7C2002A6) { // pending interrupt
-            arch_clear_pseudo_atomic_interrupted(context);
+            arch_clear_pseudo_atomic_interrupted(get_sb_vm_thread());
             arch_skip_instruction(context);
             interrupt_handle_pending(context);
             return;
@@ -616,10 +516,10 @@ sigtrap_handler(int signal, siginfo_t *siginfo, os_context_t *context)
     if (signal == SIGTRAP && handle_tls_trap(context, pc, code)) return;
 #endif
 
-    if (code == ((3 << 26) | (0x18 << 21) | (reg_NL3 << 16))||
+    if (code == ((3 << 26) | (0x18 << 21) | (reg_NL3 << 16))|| // TWI NE,$NL3,0
         /* trap instruction from do_pending_interrupt */
-        code == 0x7fe00008) {
-        arch_clear_pseudo_atomic_interrupted(context);
+        code == 0x7fe00008) { // TW T,0,0
+        arch_clear_pseudo_atomic_interrupted(get_sb_vm_thread());
         arch_skip_instruction(context);
         /* interrupt or GC was requested in PA; now we're done with the
            PA section we may as well get around to it */
@@ -694,7 +594,7 @@ ppc_flush_icache(os_vm_address_t address, os_vm_size_t length)
 void
 arch_write_linkage_table_entry(int index, void *target_addr, int datap)
 {
-  char *reloc_addr = (char*)LINKAGE_TABLE_SPACE_START + index * LINKAGE_TABLE_ENTRY_SIZE;
+  char *reloc_addr = (char*)ALIEN_LINKAGE_TABLE_SPACE_START + index * ALIEN_LINKAGE_TABLE_ENTRY_SIZE;
   if (datap) {
     *(unsigned long *)reloc_addr = (unsigned long)target_addr;
     return;
@@ -834,3 +734,19 @@ arch_write_linkage_table_entry(int index, void *target_addr, int datap)
 
   os_flush_icache((os_vm_address_t) reloc_addr, (char*) inst_ptr - reloc_addr);
 }
+
+void gcbarrier_patch_code(void* where, int nbits)
+{
+#ifdef LISP_FEATURE_64_BIT
+    int m_operand = 64 - nbits;
+    // the M field has a kooky encoding
+    int m_encoded = ((m_operand & 0x1F) << 1) | (m_operand >> 5);
+    unsigned int* pc = where;
+    unsigned int inst = *pc;
+    // .... ____ _xxx xxx_ ____ = 0x7E0;
+    //                  ^ deposit it here, in (BYTE 6 5) of the instruction.
+    *pc = (inst & ~0x7E0) | (m_encoded << 5);
+#else
+    lose("can't patch rldicl in 32-bit"); // illegal instruction
+#endif
+}
