@@ -43,7 +43,7 @@
 #include <time.h>
 
 #ifndef LISP_FEATURE_WIN32
-#include "signal.h"
+#include <signal.h>
 #endif
 
 #include "runtime.h"
@@ -51,7 +51,6 @@
 #include "globals.h"
 #include "os.h"
 #include "interr.h"
-#include "alloc.h"
 #include "interrupt.h"
 #include "arch.h"
 #include "gc.h"
@@ -323,7 +322,7 @@ parse_size_arg(char *arg, char *arg_name)
 
 char *core_string;
 
-static void print_environment(int argc, char *argv[])
+static void print_environment(int argc, char *argv[], char *envp[])
 {
     int n = 0;
     printf("; Commandline arguments:\n");
@@ -333,8 +332,8 @@ static void print_environment(int argc, char *argv[])
     }
     n = 0;
     printf(";\n; Environment:\n");
-    while (ENVIRON[n]) {
-        printf(";  %2d: \"%s\"\n", n, ENVIRON[n]);
+    while (envp[n]) {
+        printf(";  %2d: \"%s\"\n", n, envp[n]);
         ++n;
     }
 }
@@ -366,13 +365,17 @@ char *dir_name(char *path) {
 }
 
 
-extern void write_protect_immobile_space();
 struct lisp_startup_options lisp_startup_options;
 
 struct cmdline_options {
     char *core;
-    char **argv;
-    boolean disable_lossage_handler_p;
+#ifdef LISP_FEATURE_WIN32
+    wchar_t
+#else
+    char
+#endif
+     **argv;
+    bool disable_lossage_handler_p;
     int merge_core_pages;
 };
 
@@ -422,7 +425,7 @@ static int is_memsize_arg(char *argv[], int argi, int argc, int *merge_core_page
 
 static struct cmdline_options
 parse_argv(struct memsize_options memsize_options,
-           int argc, char *argv[], char *core)
+           int argc, char *argv[], char *envp[], char *core)
 {
 #ifdef LISP_FEATURE_WIN32
     wchar_t
@@ -431,13 +434,13 @@ parse_argv(struct memsize_options memsize_options,
 #endif
         **sbcl_argv = 0;
     /* other command line options */
-    boolean disable_lossage_handler_p
+    bool disable_lossage_handler_p
 #if defined(LISP_FEATURE_SB_LDB)
         = 0;
 #else
         = 1;
 #endif
-    boolean debug_environment_p = 0;
+    bool debug_environment_p = 0;
     int merge_core_pages = -1;
 
     int argi = 1;
@@ -482,7 +485,7 @@ parse_argv(struct memsize_options memsize_options,
         // Somebody who wishes this to work for #+win32 should feel free to do the same...
 #endif
     } else {
-        boolean end_runtime_options = 0;
+        bool end_runtime_options = 0;
         /* Parse our any of the command-line options that we handle from C,
          * stopping at the first one that we don't, and leave the rest */
         while (argi < argc) {
@@ -591,7 +594,7 @@ parse_argv(struct memsize_options memsize_options,
         }
     }
     if (debug_environment_p) {
-        print_environment(argc, argv);
+        print_environment(argc, argv, envp);
     }
 
     struct cmdline_options o;
@@ -629,7 +632,7 @@ initialize_lisp(int argc, char *argv[], char *envp[])
     struct memsize_options memsize_options;
     memsize_options.present_in_core = 0;
 
-    boolean have_hardwired_spaces = os_preinit(argv, envp);
+    bool have_hardwired_spaces = os_preinit(argv, envp);
 
     interrupt_init();
 #ifdef LISP_FEATURE_UNIX
@@ -657,12 +660,12 @@ initialize_lisp(int argc, char *argv[], char *envp[])
         }
     }
 
-    struct cmdline_options options = parse_argv(memsize_options, argc, argv, core);
+    struct cmdline_options options = parse_argv(memsize_options, argc, argv, envp, core);
 
     /* Align down to multiple of page_table page size, and to the appropriate
      * stack alignment. */
     dynamic_space_size &= ~(sword_t)(BACKEND_PAGE_BYTES-1);
-#ifdef LISP_FEATURE_GENCGC
+#ifdef LISP_FEATURE_GENERATIONAL
     dynamic_space_size &= ~(sword_t)(GENCGC_PAGE_BYTES-1);
 #endif
     thread_control_stack_size &= ~(sword_t)(CONTROL_STACK_ALIGNMENT_BYTES-1);
@@ -734,13 +737,19 @@ initialize_lisp(int argc, char *argv[], char *envp[])
 #endif
 
     define_var("nil", NIL, 1);
-    define_var("t", T, 1);
+    define_var("t", LISP_T, 1);
 
     if (!options.disable_lossage_handler_p)
         enable_lossage_handler();
 
     os_link_runtime();
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
+#ifdef CALLBACK_WRAPPER_TRAMPOLINE // not defined if #-sb-thread
+     // Assign the static lisp symbol's value the address of the C function
+    // of the same name. Needed when alien linkage table is relocatable.
+    extern void callback_wrapper_trampoline();
+    SYMBOL(CALLBACK_WRAPPER_TRAMPOLINE)->value = (lispobj)callback_wrapper_trampoline;
+#endif
     /* Delayed until after dynamic space has been mapped, fixups made,
      * and/or immobile-space linkage entries written,
      * since it was too soon earlier to handle write faults. */
@@ -766,8 +775,6 @@ initialize_lisp(int argc, char *argv[], char *envp[])
     core_string = core;
     posix_argv = options.argv;
 
-    FSHOW((stderr, "/funcalling initial_function=0x%lx\n",
-          (unsigned long)initial_function));
     create_main_lisp_thread(initial_function);
     return 0;
 }
