@@ -60,7 +60,12 @@
    Unix header, according to Perry E. Metzger, in a message on
    sbcl-devel dated 2004-03-29, this is the POSIXly-correct way of
    using environ: by an explicit declaration.  -- CSR, 2004-03-30 */
+#if defined __W32API_USE_DLLIMPORT__
+// Assume that if that DLLIMPORT preprocessor symbol is defined, then 'environ' is too.
+// It gets defined as '_environ' which in turn is defined as (* __p__environ())
+#else
 extern char **environ;
+#endif
 
 /*
  * stuff needed by CL:DIRECTORY and other Lisp directory operations
@@ -524,12 +529,6 @@ int s_issock(mode_t mode)
 #endif /* !LISP_FEATURE_WIN32 */
 
 #ifdef LISP_FEATURE_UNIX
-#ifdef LISP_FEATURE_DARWIN
-/* nanosleep() is not re-entrant on some versions of Darwin and is
- * reimplemented using the underlying syscalls.
- */
-int sb_nanosleep(time_t sec, int nsec);
-#else
 void sb_nanosleep(time_t sec, int nsec)
 {
     struct timespec rqtp = {sec, nsec};
@@ -537,36 +536,8 @@ void sb_nanosleep(time_t sec, int nsec)
 
     while(nanosleep(&rqtp, &rmtp) && errno == EINTR) {
         rqtp = rmtp;
-        /* The old lisp version stated
-           ;; KLUDGE: On Darwin, if an interrupt cases nanosleep to
-           ;; take longer than the requested time, the call will
-           ;; return with EINT and (unsigned)-1 seconds in the
-           ;; remainder timespec, which would cause us to enter
-           ;; nanosleep again for ~136 years. So, we check that the
-           ;; remainder time is actually decreasing.
-           ;;
-           ;; It would be neat to do this bit of defensive
-           ;; programming on all platforms, but unfortunately on
-           ;; Linux, REM can be a little higher than REQ if the
-           ;; nanosleep() call is interrupted quickly enough,
-           ;; probably due to the request being rounded up to the
-           ;; nearest HZ. This would cause the sleep to return way
-           ;; too early.
-           #+darwin
-           (let ((rem-sec (slot rem 'tv-sec))
-           (rem-nsec (slot rem 'tv-nsec)))
-           (when (or (> secs rem-sec)
-           (and (= secs rem-sec) (>= nsecs rem-nsec)))
-           ;; Update for next round.
-           (setf secs  rem-sec
-           nsecs rem-nsec)
-           t)
-
-           but the Darwin variant is implemented elsewhere
-        */
     }
 }
-#endif
 
 void sb_nanosleep_double(double seconds) {
     /* Some (which?) platforms, apparently, can't sleep more than 100
@@ -616,6 +587,18 @@ int sb_utimes(char *path, struct timeval times[2])
 {
     return utimes(path, times);
 }
+
+int sb_clock_gettime(clockid_t id, struct timespec* tp)
+{
+    return clock_gettime(id, tp);
+}
+#ifndef LISP_FEATURE_SB_THREAD
+#include <signal.h>
+int sb_sigprocmask(int how, const sigset_t *set, sigset_t *oldset)
+{
+    return sigprocmask(how, set, oldset);
+}
+#endif
 #endif
 
 #ifdef LISP_FEATURE_WIN32
